"""
Frame splitting utilities.
"""
import os
import json
import numpy
from tqdm import tqdm
from glob import glob
from concurrent.futures.thread import ThreadPoolExecutor

from stardust.rosetta.frame_tools import FrameCapture
from stardust.rosetta.structure import Slot, Input


def get_input(children, frame_number):
    attribute_input = Input(children, frame_number)
    for _ in ['frame_list', 'target']:
        if hasattr(attribute_input, _):
            delattr(attribute_input, _)
    return vars(attribute_input)


def get_slot(slot, frame_number, capture):
    if slot['type'] in FrameCapture.slot_type_list:
        attribute_slot = Slot(slot, frame_number, capture)
        if capture.get_slot_structure(attribute_slot):
            return attribute_slot
        else:
            return False
    elif slot['type'] == 'slot':
        attribute_slot = FrameCapture.get_annotation_struct(slot, 'slots')
        for slot_line in slot['slots']:
            if slot_line['type'] in FrameCapture.slot_type_list:
                slot_dict = Slot(slot_line, frame_number, capture)
                if capture.get_slot_structure(slot_dict):
                    if hasattr(slot_dict, 'capture'):
                        delattr(slot_dict, 'capture')
                    attribute_slot['slots'].append(vars(slot_dict))
        return attribute_slot


def get_slots_children(annotation, frame_number, capture: FrameCapture):
    slots_children_dict = capture.get_annotation_struct(annotation, 'slotsChildren')
    for i in annotation['slotsChildren']:
        result_slot = get_slot(i['slot'], frame_number, capture)
        if not result_slot:
            continue
        children_list = []  # Collect child annotations
        for children in i['children']:
            if children['type'] == 'input':
                result_children = get_input(children, frame_number)
            elif children['type'] == 'slot':
                result_children = get_slot(children, frame_number, capture)
            elif children['type'] == 'slotChildren':
                result_children = get_slots_children(children, frame_number, capture)
            else:
                raise ValueError(f'The children result type {children["type"]} is missing')
            children_list.append(result_children)

        if result_slot:
            if hasattr(result_slot, 'capture'):
                delattr(result_slot, 'capture')
            slots_children_dict['slotsChildren'].append({
                'slot': vars(result_slot),
                'children': children_list})
    return slots_children_dict


def get_children_only(annotation, frame_number, capture):
    children_only_dict = {
        "key": annotation['key'],
        "label": annotation['label'],
        "type": annotation['type'],
        "childrenOnly": []}
    for i in annotation['childrenOnly']:

        children_list = []  # Collect child annotations
        for children in i['children']:
            if children['type'] == 'input':
                result_children = get_input(children, frame_number)
            elif children['type'] == 'slot':
                result_children = get_slot(children, frame_number, capture)
            elif children['type'] == 'slotChildren':
                result_children = get_slots_children(children, frame_number, capture)
            else:
                result_children = {}
                print(f"result_children {children['type']} attribute missing")
            children_list.append(result_children)
        children_only_dict['childrenOnly'].append({
            'id': i['id'],
            'label': i['label'],
            'children': children_list})
    return children_only_dict


def frames_split_task(file_path):
    try:
        frame = FrameCapture(file_path)
    except AssertionError:
        return True
    else:
        for frame_number in range(frame.attachment_length):
            with open(file_path.replace('.json', f'_{str(frame_number).rjust(4, "0")}.json'), "w") as fw:
                frame_data = {
                    "projectId": frame.projectId,
                    "datasetId": frame.datasetId,
                    "poolId": frame.poolId,
                    "taskId": frame.taskId,
                    "status": frame.status,
                    "taskParams": {
                        "record": {
                            "attachmentType": frame.attachment_type.replace("_SEQUENCE", ""),
                            "attachment": frame.attachment[frame_number],
                            "metadata": frame.metadata
                        },
                        "operators": frame.operators
                    }
                }
                annotation_frame_data = []
                for annotation in frame.annotations:
                    try:
                        if annotation['type'] == 'slotChildren':
                            _ = get_slots_children(annotation, frame_number, frame)
                        elif annotation['type'] == 'childrenOnly':
                            _ = get_children_only(annotation, frame_number, frame)
                        elif annotation['type'] == 'input':
                            _ = get_input(annotation, frame_number)
                        elif annotation['type'] == 'slot':
                            _ = get_slot(annotation, frame_number, frame)
                        else:
                            raise ValueError(f'The annotation result type {annotation["type"]} is missing')
                    except numpy.core._exceptions.UFuncTypeError:
                        print(f'{frame_number} encountered invalid boxes')
                    except ValueError as e:
                        print(f'{frame_number} {e}')
                    else:
                        annotation_frame_data.append(_)
                frame_data['result'] = {
                    'annotations': annotation_frame_data,
                }
                fw.write(json.dumps(frame_data, ensure_ascii=False))
        os.remove(file_path)
        return True


def frames_split(input_path):
    file_paths = [file for file in glob(f'{input_path}/**', recursive=True) if file.endswith('.json')]
    bar = tqdm(total=len(file_paths))
    with ThreadPoolExecutor(max_workers=4) as executor:
        executors = executor.map(frames_split_task, file_paths)
        for exe in executors:
            if exe:
                bar.update(1)


# Each instance is split frame-by-frame
def func_ms(ms_data: "ms_data"):
    """
    Iterate over all instances; each may contain a sequence of frames.

    Args:
        ms_data:

    Returns:
        All frames for each instance

    """

    result_lst = list()
    for instance in ms_data['data']['data']:
        if "SEQUENCE" not in instance['record']['attachmentType']:
            result_lst.append([{
                "poolId": None,
                "taskId": None,
                "datasetId": instance['datasetId'],
                "dataInstanceId": instance['dataInstanceId'],
                "projectId": instance['projectId'],
                "algoId": instance['algoId'],
                "versionNum": instance['versionNum'],
                "datasetName": instance['datasetName'],
                "taskParams": {
                    "record": {
                        "attachmentType": instance['record']['attachmentType'].replace("_SEQUENCE", ""),
                        "attachment": instance['record']['attachment'],
                        "metadata": instance['record']['metadata']
                    }
                },
                "result": {
                    "annotations": instance['annotation']
                }
            }, ])

        else:
            frame_numbers = len(instance['record']['attachment'])
            current_instance_data_lst = list()  # Container for all frames of the current instance

            # Iterate per frame
            for frame_number in range(frame_numbers):
                ret_data = {
                    "poolId": None,
                    "taskId": None,
                    "datasetId": instance['datasetId'],
                    "dataInstanceId": instance['dataInstanceId'],
                    "projectId": instance['projectId'],
                    "algoId": instance['algoId'],
                    "versionNum": instance['versionNum'],
                    "datasetName": instance['datasetName'],
                    "taskParams": {
                        "record": {
                            "attachmentType": instance['record']['attachmentType'].replace("_SEQUENCE", ""),
                            "attachment": instance['record']['attachment'][frame_number],
                            "metadata": instance['record']['metadata']
                        }
                    },
                    "result": {
                        "annotations": list()
                    }
                }
                annotation_frame_data = []
                for annotation in instance['annotation']:
                    try:
                        if annotation['type'] == 'slotChildren':
                            _ = get_slots_children(annotation, frame_number, FrameCapture)
                        elif annotation['type'] == 'childrenOnly':
                            _ = get_children_only(annotation, frame_number, FrameCapture)
                        elif annotation['type'] == 'input':
                            _ = get_input(annotation, frame_number)
                        elif annotation['type'] == 'slot':
                            _ = get_slot(annotation, frame_number, frame)
                        else:
                            raise ValueError(f'The annotation result type {annotation["type"]} is missing')
                    except numpy.core._exceptions.UFuncTypeError:
                        print(f'{frame_number} encountered invalid boxes')
                    except ValueError as e:
                        print(f'{frame_number} {e}')
                    else:
                        annotation_frame_data.append(_)
                        break
                ret_data['result']['annotations'] = annotation_frame_data
                current_instance_data_lst.append(ret_data)

            result_lst.append(current_instance_data_lst)

    return result_lst


if __name__ == '__main__':
    for pid in (1953, 2011, 2026):
        # pid = 1337
        dir_path = f'/Users/mac/Documents/pyproject/Chacer/rosetta/{pid}/'

        frames_split(dir_path)



import math
from copy import deepcopy
from typing import List, Dict

import open3d as o3d
import numpy as np
from scipy.spatial.transform import Rotation as R


class Matrix4(object):
    def __init__(self, box: tuple):
        self.box = box
        self.te = deepcopy(np.array(self.box).reshape(4, 4))
        self._determinant = np.linalg.det(self.te.T)
        self.sx = np.sqrt(np.sum(self.te[:1, :3].flatten() ** 2))
        self.sy = np.sqrt(np.sum(self.te[1:2, :3].flatten() ** 2))
        self.sz = np.sqrt(np.sum(self.te[2:3, :3].flatten() ** 2))
        if self._determinant < 0:
            self.sx = -self.sx
        else:
            self.sx = self.sx

    def to_position(self):
        return dict(zip('xyz', self.te.T[:3, 3:].flatten().tolist()))

    def to_rotation(self):
        inv_sx = 1 / self.sx
        inv_sy = 1 / self.sy
        inv_sz = 1 / self.sz

        _ = np.reshape([inv_sx, inv_sy, inv_sz], [-1, 1])
        m = np.multiply(self.te[:3, :3], _)

        euler = R.from_matrix(m.T).as_euler("XYZ", degrees=False).tolist()
        return dict(zip('xyz', euler))

    def to_scale(self):
        return {
            "length": self.sx,
            "width": self.sy,
            "height": self.sz
        }

    @property
    def box_info(self):
        return {
            "position": self.to_position(),
            "scale": self.to_scale(),
            "rotation": self.to_rotation()
        }

    @classmethod
    def get_transform_matrix(cls, box: Dict):
        position = box['position']
        scale = box['scale']
        rotation = [box['rotation'][key] for key in ['x', 'y', 'z']]

        # Compute the rotation matrix
        rx, ry, rz = rotation
        cx, cy, cz = np.cos(rx), np.cos(ry), np.cos(rz)
        sx, sy, sz = np.sin(rx), np.sin(ry), np.sin(rz)
        rotation_matrix = np.array([[cy * cz, -cy * sz, sy, 0],
                                    [cx * sz + cz * sx * sy, cx * cz - sx * sy * sz, -cy * sx, 0],
                                    [sx * sz - cx * cz * sy, cx * sz * sx + cz * sy, cx * cy, 0],
                                    [0, 0, 0, 1]])

        # Compute the scaling matrix
        scale_matrix = np.eye(4)
        scale_matrix[0, 0] = scale['length']
        scale_matrix[1, 1] = scale['width']
        scale_matrix[2, 2] = scale['height']

        # Compute the translation matrix
        translation_matrix = np.eye(4)
        translation_matrix[:3, 3] = [position['x'], position['y'], position['z']]

        # Compose the full transform matrix
        transform_matrix = np.matmul(translation_matrix, np.matmul(rotation_matrix, scale_matrix))
        transform_matrix = transform_matrix.T

        # Flatten to a 1D list
        transform_list = transform_matrix.flatten().tolist()

        return transform_list

    @classmethod
    def get_points_matrix(cls, box_info: Dict, rt: np.matrix):
        # Convert point P into homogeneous coordinates
        box_info = [box_info['x'], box_info['y'], box_info['z']]
        p_homogeneous = np.hstack([box_info, 1])
        # Apply the affine transform
        p_homogeneous_transformed = np.dot(rt, p_homogeneous)
        p_transformed = p_homogeneous_transformed[:3] / p_homogeneous_transformed[3]

        # Return the transformed coordinates
        return dict(zip("xyz", p_transformed))

    @classmethod
    def get_3dbox_matrix(cls, box_info: Dict, rt: np.matrix):
        center = [box_info['position']['x'], box_info['position']['y'], box_info['position']['z']]  # Center point
        extent = [box_info['scale']['height'], box_info['scale']['width'], box_info['scale']['depth']]  # Extent per axis
        theta = [box_info['rotation']['x'], box_info['rotation']['y'], box_info['rotation']['z']]  # Euler angles
        # Convert Euler angles to rotation matrix

        r = R.from_euler('xyz', theta)
        rot_mat = r.as_matrix()
        bbox = o3d.geometry.OrientedBoundingBox(center=center, R=rot_mat, extent=extent)
        r1 = np.array(rt)
        bbox.rotate(r1[:3, :3])
        bbox.translate(r1[:3, 3:].reshape(-1))
        euler = R.from_matrix(bbox.R.copy()).as_euler("XYZ", degrees=False).tolist()
        # Package center, dimensions, and Euler angles

        box = {
            'position': {
                'x': bbox.center[0],
                'y': bbox.center[1],
                'z': bbox.center[2]},
            'scale': {
                'height': bbox.extent[0],
                'width': bbox.extent[1],
                'depth': bbox.extent[2]},
            'rotation': {
                'x': euler[0],
                'y': euler[1],
                'z': euler[2],
                '_order': 'XYZ'}}

        return box

    @classmethod
    def euler_to_rotation_matrix(cls, euler_angles):
        roll, pitch, yaw = euler_angles
        x = np.array([[1, 0, 0],
                      [0, np.cos(roll), -np.sin(roll)],
                      [0, np.sin(roll), np.cos(roll)]])
        y = np.array([[np.cos(pitch), 0, np.sin(pitch)],
                      [0, 1, 0],
                      [-np.sin(pitch), 0, np.cos(pitch)]])
        z = np.array([[np.cos(yaw), -np.sin(yaw), 0],
                      [np.sin(yaw), np.cos(yaw), 0],
                      [0, 0, 1]])
        r = np.dot(z, np.dot(y, x))
        return r

    @classmethod
    def get_3dbox_same_frame(cls, box_info, coordinate):
        center = [box_info['position']['x'], box_info['position']['y'], box_info['position']['z'], 1]  # Center point
        theta = [0, 0, box_info['rotation']['z']]  # Euler angles
        # Convert Euler angles to rotation matrix

        center = np.array(center)
        center_new = np.dot(center, coordinate.T).tolist()[0]

        _ = cls.euler_to_rotation_matrix(theta)
        r = np.dot(_, coordinate[:3, :3])

        _, _, rz_new = R.from_matrix(r).as_euler('xyz')

        box = {
            'position': {
                'x': center_new[0],
                'y': center_new[1],
                'z': center_new[2]},
            'scale': box_info['scale'],
            'rotation': {
                'x': 0,
                'y': 0,
                'z': rz_new,
                '_order': 'XYZ'}}

        return box

    @staticmethod
    def to_matrix4d(cx, cy, cz, dx, dy, dz, heading):
        """
        7 number Matrix4
        :param cx:
        :param cy:
        :param cz:
        :param dx: Length in the x direction   width
        :param dy: Length in the y direction   height
        :param dz: Length in the z direction
        :param heading: yaw angle; 0 on the positive x-axis, positive counter-clockwise
        :return:
        """
        alpha, beta, gamma = 0, 0, heading  # TODO: pitch, roll, yaw
        x = math.cos(alpha / 2) * math.sin(beta / 2) * math.cos(gamma / 2) + math.sin(alpha / 2) * math.cos(
            beta / 2) * math.sin(gamma / 2)
        y = math.sin(alpha / 2) * math.cos(beta / 2) * math.cos(gamma / 2) - math.cos(alpha / 2) * math.sin(
            beta / 2) * math.sin(gamma / 2)
        z = -math.sin(alpha / 2) * math.sin(beta / 2) * math.cos(gamma / 2) + math.cos(alpha / 2) * math.cos(
            beta / 2) * math.sin(gamma / 2)
        w = math.cos(alpha / 2) * math.cos(beta / 2) * math.cos(gamma / 2) + math.sin(alpha / 2) * math.sin(
            beta / 2) * math.sin(gamma / 2)

        x2 = x + x
        y2 = y + y
        z2 = z + z

        xx = x * x2
        xy = x * y2
        xz = x * z2

        yy = y * y2
        yz = y * z2
        zz = z * z2

        wx = w * x2
        wy = w * y2
        wz = w * z2

        sx, sy, sz = dy, dx, dz

        te = [0] * 16
        te[0] = (1 - (yy + zz)) * sx
        te[1] = (xy + wz) * sx
        te[2] = (xz - wy) * sx
        te[3] = 0

        te[4] = (xy - wz) * sy
        te[5] = (1 - (xx + zz)) * sy
        te[6] = (yz + wx) * sy
        te[7] = 0

        te[8] = (xz + wy) * sz
        te[9] = (yz - wx) * sz
        te[10] = (1 - (xx + yy)) * sz
        te[11] = 0

        te[12] = cx
        te[13] = cy
        te[14] = cz
        te[15] = 1

        return te


if __name__ == '__main__':
    res = Matrix4((-1.8629903392376739, -0.07255098381334309, 0, 0, 0.180460797718889, -4.633937475284563, 0, 0, 0, 0,
                   1.630792052527909, 0, -11.792962546858156, -12.32071494109367, 0.963595499026534, 1)).box_info
    print(res)
    res1 = Matrix4.get_transform_matrix(res)
    print(res1)



import json
import uuid
import time
import os

current_path = os.path.dirname(__file__)


# TODO: integrate with ossutil helpers
def ToAnnotationTool(data_path: str, task_type: str):
    """
    Visual result of algorithm with AnnotationTool

    Args:
        data_path: str
            The path of algorithm output which should be a json file
        task_type: str
            The type of visual task, the task can be chosen from 'object_detection', 'lane_detection' and 'tracking' for now

    Returns:
        str: 
            URL of AnnotationTool

    Examples:
        .. code-block:: python

            from stardust.visualization.visual import ToAnnotationTool
            vis_url = ToAnnotationTool('tracking_test.json', 'tracking')
    """
    assert task_type in ['object_detection', 'lane_detection', 'tracking']
    assert data_path.endswith('.json')
    with open(data_path, 'r') as f:
        input_data = json.load(f)

    log_time = time.strftime('%Y%m%d-%H%M%S', time.localtime())
    base_path = os.path.join(current_path, 'base', f'{task_type}.json')
    with open(base_path, 'r') as f:
        output = json.load(f)

    output['taskParams']['value']['record']['metadata'] = {
        'uniqueIdentifier': str(uuid.uuid4())
    }

    output['taskParams']['value']['record']['attachment'] = input_data['attachment']

    if task_type == 'object_detection':
        result_type = input_data['results']['type']
        annotations = input_data['results'][result_type]
        for obj_info in annotations:
            slot = obj_info['slot']
            children = obj_info['children']
            output['taskResult']['annotations'][0]['slotsChildren'].append(
                {
                    "slot":
                        {
                            "id": slot['id'],
                            "box": slot['box'],
                            "type": "box3d",
                            "label": "Point Cloud Box"},
                    "children": [{
                        "key": "box2d-[b274c]",
                        "type": "slot",
                        "label": "Mapped Bounding Box",
                        "slots": []}]}
            )
    elif task_type == 'lane_detection':
        result_type = input_data['results']['type']
        annotations = input_data['results'][result_type]
        for obj_info in annotations:
            output['taskResult']['annotations'][0]['slots'].append(
                {
                    "id": obj_info['id'],
                    "vertices": obj_info['vertices'],
                    "type": "line",
                    "label": "lane line"}
            )
    elif task_type == 'tracking':
        fallback_box_key = '3D\u6846'
        box_key = '3D box' if '3D box' in input_data['results'] else fallback_box_key
        fallback_rect_key = '\u77e9\u5f62\u6846'
        rect_key = 'Mapped Bounding Box' if 'Mapped Bounding Box' in input_data['results'].get(box_key, {}) else fallback_rect_key
        result_type = input_data['results'][box_key]['type']
        box3d = input_data['results'][box_key][result_type]
        for obj_info in box3d:
            slot = obj_info['slot']
            children = obj_info['children']
            output['taskResult']['annotations'][0]['slotsChildren'].append(
                {
                    "slot":
                        {
                            "id": slot['id'],
                            "box": slot['box'],
                            "timeSeries": slot['timeSeries'],
                            "type": "box3d",
                            "label": "Point Cloud Box"}}
            )
            if rect_key in children:
                for child in children[rect_key]['slot']:
                    child['label'] = 'Mapped Bounding Box'
                output['taskResult']['annotations'][0]['slotsChildren'][-1].update(children=[{
                    "key": "box2d-[b274c]",
                    "type": "slot",
                    "label": "Mapped Bounding Box",
                    "slots": children[rect_key]['slot']}])
            else:
                output['taskResult']['annotations'][0]['slotsChildren'][-1].update(children=[{}])
    os.makedirs(f'./results/{task_type}', exist_ok=True)
    local_path = f'./results/{task_type}/{log_time}.json'
    with open(local_path, 'w') as f:
        json.dump(output, f, ensure_ascii=False)
    oss_path = f'oss://stardust-data/Clients/J-ceshi/path-test/Production/666/{task_type}/{log_time}.json'
    vis_path = oss_path[oss_path.find('Clients'):]
    vis_url = f'https://dev.rosettalab.top/annotate#fromLink/https://stardust-data.rosettalab.top/{vis_path}'
    print(vis_url)
    os.system(f'ossutil cp -rf {local_path} {oss_path} -e oss-cn-hangzhou.aliyuncs.com')
    return vis_url



from .base_annotation import AnnotationObject


class Audio(AnnotationObject):
    def __init__(self, start, end, **kwargs):
        """
        This class defines the structure of video label task

        Args:
            start: float
                The start time of annotation for this audio
            end: float
                The end time of annotation for this audio

        Returns:
            Audio instance

        Examples:
            .. code-block:: python

                audio = Audio(start=0.57, 12.13)
        """
        assert isinstance(start, float)
        assert isinstance(end, float)
        self.start = start
        self.end = end
        AnnotationObject.__init__(self, **kwargs)



"""
    Get a token for the rosetta platform
"""
__all__ = ['RosettaToken']
import time
import random
import json

import requests

from stardust.config import RosConfig, RosUser


class Auth:
    def __init__(self):
        self.username: str = RosUser.username
        self.password: str = RosUser.password


class RosettaToken(Auth):
    def __init__(self, env: str):
        super().__init__()
        self.login_url: str = f'{env}/rosetta-service/user/login'
        self.session_id: str = ''
        self.authorize: str = ''

    def get_authorize(self):
        headers = {
            "accept": "application/json, text/plain, */*",
            "content-type": "application/json",
        }
        data = {
            "username": self.username,
            "password": self.password
        }
        data = json.dumps(data, separators=(',', ':'))
        response = requests.post(self.login_url, headers=headers, data=data)
        self.authorize = response.json()['data']['tokenValue']

    def get_headers(self):
        def generate_string():
            r = 20
            n = [''] * r
            o = list(str(int(time.time() * 1000))[::-1])
            for i in range(r):
                e = random.randint(0, 35)
                t = str(e if e < 26 else chr(e + 87))
                n[i] = t if e % 3 else t.upper()
            for i in range(8):
                n.insert(3 * i + 2, o[i])
            return ''.join(n[::-1])

        self.session_id = generate_string()
        self.get_authorize()
        headers = {
            "authority": "server.rosettalab.top",
            "accept": "application/json, text/plain, */*",
            "accept-language": "zh-CN",
            "authorize": self.authorize,
            "content-type": "application/json",
            "eagleeye-sessionid": self.session_id,
            "origin": "https://rosettalab.top",
            "referer": "https://rosettalab.top/",
            "sec-ch-ua": "\"Google Chrome\";v=\"111\", \"Not(A:Brand\";v=\"8\", \"Chromium\";v=\"111\"",
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": "\"macOS\"",
            "sec-fetch-dest": "empty",
            "sec-fetch-mode": "cors",
            "sec-fetch-site": "same-site",
            "user-agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/111.0.0.0 Safari/537.36"
        }
        return headers


if __name__ == '__main__':
    exp = RosettaToken(RosConfig.env_prod)
    exp.get_authorize()
    print(exp.get_headers())



from typing import Dict, Iterable, List, Optional
from numbers import Number
import re
from pathlib import Path
import json
import glob
import shutil

from stardust.components.annotations import Box2D, Box3D, Input, Line, Point, Polygon, Text
from stardust.components.attachment import Image, PointCloud
from stardust.components.frame import Annotation, Frame, Media, Prediction, TaskInfo
from stardust.rosetta.rosetta_data import RosettaData as Export
from stardust.ms.ms import MS

__all__ = [
    "read_rosetta", "read_ms"
]


def _serialize(obj):
    dic = dict()
    if hasattr(obj, "__dict__"):
        for key, val in obj.__dict__.items():
            if str(key).startswith("__") and str(key).endswith("__"):
                continue
            if str(key) in ("id", "shapely_polygon"):
                continue
            dic[key] = _serialize(val)
    elif isinstance(obj, (str, Number)):
        return obj
    elif isinstance(obj, Dict):
        for key, val in obj.items():
            dic[key] = _serialize(val)
    elif isinstance(obj, Iterable):
        return [_serialize(val) for val in obj]
    return dic


class Convertor:
    def __init__(self) -> None:
        self.project_id = None
        self.input_path = None
        self.pattern = re.compile(r'_(\d+)\.json')

    def yield_rosetta_json(self):
        """
        yield json file from local path which is download from rosetta

        Returns:

        """

        if not (rosetta_json_lst := glob.glob(os.path.join(self.input_path, f"{self.project_id}/json/**/*.json"))):
            rosetta_json_lst = glob.glob(os.path.join(self.input_path, f"{self.project_id}/json/*.json"))

        for file in rosetta_json_lst:
            try:
                with open(file, 'r', encoding='utf-8') as jf:
                    _ = json.load(jf)
                    self.file_name = str(file)
                    __ = Path(file).stem.rsplit("_", 1)[1]
                    if self.pattern.search(str(file)):
                        self.frame_num = int(Path(file).stem.rsplit("_", 1)[1])
                    else:
                        self.frame_num = 0
                    yield _
            except json.decoder.JSONDecodeError as e:
                print(e)
                continue
            except Exception as e:
                raise e

    def factory_children(self, children: "rosetta children",
                         parent_id: str = None):
        """
        Start processing operation subentries, recursion

        Args:
            children:
                rosetta children
            slot_id:
                rosetta slot id

        Returns:
            Dictionary, where key is the id of children and value is the type
        """
        children_lst = dict()
        label_kind = None
        for child in children:
            if child['type'] == "input":
                input_obj = Input.gen_input(child, parent_id)
                children_lst[input_obj.id] = "input"
                if isinstance(input_obj.name, str) and input_obj.name.lower() in ("type", "category"):
                    label_kind = input_obj.value
                self.label_result.input_lst.setdefault(input_obj.id, input_obj)
            elif child['type'] == "slotChildren":
                self.factory_slot_children(child['slotsChildren'], parent_id, children_lst)
            elif child['type'] == "slot":
                self.factory_slots(child['slots'], parent_id, children_lst)
            else:
                raise ValueError
        return children_lst, label_kind

    def factory_slot(self, slot: "rosetta slot",
                     children_lst: Dict = None,
                     parent_id: str = None,
                     label_kind: str = None,
                     team_id: int = None) -> Optional[str]:
        """
        Processing label box

        Args:
            slot:
                rosetta slot
            children_lst:
                rosetta children
            parent_id:
                Upper floor ID
        Returns:
            str， id
        """
        obj_type = slot['type']
        if obj_type == "box3d":  # 3D bounding box
            box3d = Box3D.gen_box3d(slot, children_lst, parent_id, label_kind, team_id=team_id)
            if not box3d:
                return
            self.label_result.box3d_lst.setdefault(box3d.id, box3d)
            return box3d.id
        elif obj_type == "box2d":  # 2D bounding box
            box2d = Box2D.gen_box(slot, children_lst, parent_id, label_kind, team_id=team_id)
            if not box2d:
                return
            self.label_result.box2d_lst.setdefault(box2d.id, box2d)
            return box2d.id
        elif obj_type == "cuboid":  # Mapped cuboid
            pass
        elif obj_type == "line":  # 2D line
            line = Line.gen_line(slot, children_lst, parent_id, label_kind, team_id=team_id)
            if not line:
                return
            self.label_result.line_lst.setdefault(line.id, line)
            return line.id
        elif obj_type == "polygon":  # Polygon
            polygon = Polygon.gen_polygon(slot, children_lst, parent_id, label_kind, team_id=team_id)
            if not polygon:
                return
            self.label_result.polygon_lst.setdefault(polygon.id, polygon)
            return polygon.id
        elif obj_type == "point":  # 2D keypoint
            point = Point.gen_point(slot, children_lst, parent_id, label_kind, team_id=team_id)
            if not point:
                return
            self.label_result.key_point_lst.setdefault(point.id, point)
        elif obj_type == "text":  # Text annotation
            text = Text.gen_text(slot, children_lst, parent_id, label_kind, team_id=team_id)
            if not text:
                return
            self.label_result.text_lst.setdefault(text.id, text)
        else:
            print(f'Unsupported structure: {obj_type}')
            pass
            # raise ValueError('Unsupported structure')

    def factory_slots(self, slot_lst: "rosetta slots json",
                      parent_id: str = None,
                      children_lst: Dict = {}) -> None:
        """
        Start processing the tag box list

        Args:
            slot_lst:
                rosetta slot
            parent_id:
                Upper floor ID
            children_lst:
                rosetta children

        Returns:

        """
        for slot in slot_lst:
            slot_id = self.factory_slot(slot, list(), parent_id=parent_id, team_id=slot.get("teamId", None))
            if slot_id:
                children_lst[slot_id]: slot['type']

    def factory_slot_children(self, sc_lst: "slotsChildren or slots",
                              parent_id: str = None,
                              children_lst: Dict = {}) -> None:
        """
        Start processing the annotation instance

        Args:
            children_lst:
                rosetta children
            sc_lst:
                rosetta slot
            parent_id:
                Upper floor ID

        Returns:

        """
        for sc in sc_lst:
            try:
                slot_id = sc['slot']['id']
                if sc.get("children"):
                    _children_lst, label_kind = self.factory_children(sc['children'], slot_id)
                    self.factory_slot(sc['slot'], _children_lst, parent_id, label_kind,
                                      team_id=sc['slot'].get("teamId", None))
                    if slot_id:
                        children_lst[slot_id] = sc['slot']['type']
                else:
                    self.factory_slot(sc['slot'], list(), parent_id, team_id=sc['slot'].get("teamId", None))
            except Exception as e:
                raise e
                continue

    def factory_anns(self, ann_lst: List) -> None:
        """
        Annotated configuration

        Args:
            ann_lst:
                Annotated result list
        Returns:

        """
        for ann in ann_lst:
            ann_type = ann['type']
            if ann_type == "input":
                input = Input.gen_input(ann)
                self.label_result.input_lst.setdefault(input.id, input)
            elif ann_type == "slotChildren":
                self.factory_slot_children(ann['slotsChildren'])
            elif ann_type == "slot":
                self.factory_slots(ann['slots'])
            elif ann_type == "childrenOnly":
                pass

    def factory_label_file(self, label_type: str,
                           label_file,
                           **kwargs) -> Media:
        """
        Start processing annotation file, single frame

        Args:
            label_type:
                Annotated type, image, point cloud
            label_file:
                Mark file
        Returns:
                a Media instance
        """
        if "POINTCLOUD" in label_type:
            pcd_url = label_file['url'] if isinstance(label_file, Dict) else label_file
            meta_point = PointCloud(
                uri=pcd_url,
                name=None,
                file_path=Path(pcd_url)
            )
            meta_images = list()
            if "imageSources" in label_file:
                for image in label_file['imageSources']:
                    meta_images.append(Image.gen_image(image))
            media = Media(point_cloud=meta_point, image=meta_images)
        elif "IMAGE" in label_type:
            image_url = label_file['url'] if isinstance(label_file, Dict) else label_file
            media = Media(
                point_cloud=None,
                image=Image(uri=image_url,
                            width=kwargs['size'].get("width") if kwargs.get("size") else None,
                            height=kwargs['size'].get("height") if kwargs.get("size") else None
                            )
            )
        else:
            media = {}

        return media

    def factory_rosetta(self, data: Dict) -> Frame:
        """
        Start processing, all single frame data

        Args:
            data: rosetta data

        Returns:
            a Frame instance

        """
        label_file = data['taskParams']['record']['attachment']
        label_type = data['taskParams']['record']['attachmentType']

        # Rosetta metadata
        self.task_info = TaskInfo(
            task_id=data['taskId'],
            project_id=data['projectId'],
            pool_id=data['poolId'],
            frame_num=self.frame_num
        )

        # Prepare annotation file metadata
        self.media = self.factory_label_file(label_type=label_type,
                                             label_file=label_file,
                                             size=data['taskParams']['record']['metadata'].get("size") \
                                                 if data['taskParams']['record'].get('metadata') else None
                                             )

        # Human annotations
        self.label_result = Annotation()
        # Parse annotation results
        self.factory_anns(data['result']['annotations'])
        annotation = self.label_result

        # Model-generated predictions
        self.label_result = Prediction()

        # Process pre-annotations if present
        if not data['taskParams']['record']['metadata']:
            # return None  # Kept for legacy compatibility consideration
            prediction = None
        else:
            if pres_data := data['taskParams']['record']['metadata'].get("preprocessedData"):
                if pres_data.get("annotations"):
                    self.factory_anns(pres_data['annotations'])
            prediction = self.label_result

        return Frame(media=self.media,
                     task_info=self.task_info,
                     annotation=annotation,
                     prediction=prediction
                     )

    def convert_rosetta(self) -> Dict:
        """
        Start to do the conversion work, yield a single frame of data

        Returns:

        """
        for json_data in self.yield_rosetta_json():
            yield self.factory_rosetta(json_data)

    def export_rosetta(self, pool_lst=[], split_name='old', env_name='top') -> None:
        """
        export rosetta project data

        Args:
            pool_lst:
                A list of pools to export
        Returns:

        """
        Export(self.project_id, self.input_path, env_name).export(pool_lst, split_name=split_name)

    def read_rosetta(self, project_id,
                     input_path,
                     pool_lst,
                     export_type=None,
                     **kwargs):
        """
        start to convert data, your choice can be: SDK、 json，default to be SDK，
        if you want to export other formats, please refer to stardust.conversion

        Args:
            project_id:
                rosetta project
            input_path:
                path to store json data exported from Rosetta
            pool_lst:
                pool list from your rosetta project
            export_type:
                expected export data type，SDK or json

        Returns:

        """

        assert project_id and isinstance(project_id, int)

        self.project_id = project_id
        self.input_path = input_path

        # Export data from Rosetta
        if os.path.exists(os.path.join(self.input_path, str(self.project_id), "json")):
            print("Start to fetch annotation results")
            shutil.rmtree(os.path.join(self.input_path, str(self.project_id)))

        self.export_rosetta(pool_lst, split_name=kwargs.get("split"), env_name=kwargs.get("env_name"))

        # Yield results as a generator
        for frame_obj in self.convert_rosetta():
            if not frame_obj:
                continue
            if export_type == "json":
                yield _serialize(frame_obj)
            else:
                yield frame_obj

    def read_ms(self, export_type=None, **kwargs):
        """
        Read data from MorningStar.

        Args:
            export_type:
                Desired export type
            ms_data:
                MorningStar response payload

        Returns:
                Frame objects

        Known gaps:
            taskinfo missing
            annotation file missing
            relationship between frames and annotation file
        """
        # Fetch dataset from MorningStar
        ms_data = MS().export_dataset(**kwargs)
        assert ms_data

        # Each dataset yields instances that are processed sequentially
        for instance_lst in ms_data:
            for index, instance in enumerate(instance_lst):
                self.frame_num = index
                res = self.factory_rosetta(instance)
                if export_type == "json":
                    yield _serialize(res)
                else:
                    yield res


def read_rosetta(
        project_id: int = None,
        input_path: str = None,
        pool_lst: Optional[List[int]] = None,
        export_type=None,
        **kwargs):
    """
    Start converting data. Options include SDK and json outputs (default SDK).
    For additional formats, see stardust.conversion.

    Args:
        project_id:
            Rosetta project identifier
        input_path:
            Path used to store JSON exported from Rosetta
        pool_lst:
            Pool list from your Rosetta project
        export_type:
            Expected output type: SDK or json

    Returns:

    """
    assert project_id and input_path
    return Convertor().read_rosetta(project_id, input_path, pool_lst, export_type, **kwargs)


def read_ms(*args, **kwargs):
    """
    Export MorningStar data

    Args:
        dataset_id:
            Data set ID
        version_num:
            Version number
        slice_id:
            Slice ID
        page_no:
            Slice paging
        page_size:
            Amount of data per page

    Returns:
        generator object

    Examples:
        .. code-block:: python

            from stardust.utils.convert import read_ms
            from stardust.convertion.to_pandaset import to_pandaset

            # Derived data
            gen_data = read_ms(
                dataset_id=351787480925605888,
                version_num=18
            )
            # scale format
            to_pandaset(gen_data, export_path:"Input save directory")


    """
    return Convertor().read_ms(*args, **kwargs)


if __name__ == '__main__':
    current_dir = os.getcwd()
    base_dir = os.path.dirname(os.path.dirname(current_dir))

    project_id = 1354
    pool_lst = [33750, ]
    # Input path is the directory where Rosetta JSON files are stored
    input_path = f"{base_dir}/data/"

    os.makedirs(input_path, exist_ok=True)

    gen_data = read_rosetta(project_id=project_id,
                            input_path=input_path,
                            pool_lst=pool_lst,
                            export_type="json"
                            )

    os.makedirs(os.path.join(input_path, "output"), exist_ok=True)
    # with open(f"{os.path.join(input_path, f'{project_id}', 'output')}/{project_id}.json", 'w', encoding='utf-8') as f:
    from pprint import pprint

    for data in gen_data:
        # pprint(len(data['annotation']['box3d_lst']))
        pass
        # json.dump(data, f, ensure_ascii=False, indent=2)
        # break
        # load_dataset(data, "/Users/mac/Desktop/1835")
        # break



import json
import os
import shutil
import tempfile
import unittest
from pathlib import Path
from unittest import mock

from stardust.visualization.visual import ToAnnotationTool


FIXTURE_DIR = Path(__file__).resolve().parents[1] / "demo" / "visualization_demo"


class VisualizationSamplesTest(unittest.TestCase):
    def setUp(self) -> None:
        self._original_cwd = os.getcwd()
        self._temp_dir = tempfile.TemporaryDirectory()
        os.chdir(self._temp_dir.name)

    def tearDown(self) -> None:
        os.chdir(self._original_cwd)
        self._temp_dir.cleanup()

    def _copy_fixture(self, filename: str) -> str:
        destination = Path(self._temp_dir.name) / filename
        destination.write_text((FIXTURE_DIR / filename).read_text(encoding="utf-8"), encoding="utf-8")
        return str(destination)

    @mock.patch("os.system", return_value=0)
    def test_object_detection_fixture_generates_annotation_payload(self, mock_system: mock.MagicMock) -> None:
        data_path = self._copy_fixture("object_detection.json")
        with mock.patch("time.strftime", return_value="20240101-000000"):
            url = ToAnnotationTool(data_path, "object_detection")

        self.assertTrue(url.endswith("/object_detection/20240101-000000.json"))

        output_path = Path("results/object_detection/20240101-000000.json")
        self.assertTrue(output_path.exists())

        result_payload = json.loads(output_path.read_text(encoding="utf-8"))
        source_payload = json.loads(Path(data_path).read_text(encoding="utf-8"))
        self.assertEqual(result_payload["taskParams"]["value"]["record"]["attachment"], source_payload["attachment"])
        mock_system.assert_called_once()


if __name__ == "__main__":  # pragma: no cover
    unittest.main()



"""
Common helpers for frame decomposition.
"""
import json
import math
from stardust.components.ego.ego import Ego


class FrameCapture:
    sequence_type_list = ["IMAGE_SEQUENCE", "IMAGE_SET_SEQUENCE", "POINTCLOUD_SEQUENCE",
                          "POINTCLOUD_SET_SEQUENCE"]
    slot_type_list = ['box3d', 'box2d', 'cuboid', 'polygon3d', 'point3d', 'line3d', 'line', 'splines', 'point',
                      'polygon']
    slot_attr_list = ['matrix', 'plane', 'vertices', 'points', 'point']

    def __init__(self, path):
        with open(path) as fp:
            _ = json.load(fp)
            record = _['taskParams']['record']
            self.attachment_type = record['attachmentType']
            assert self.attachment_type in self.sequence_type_list
            self.projectId = _['projectId']
            self.datasetId = _['datasetId']
            self.poolId = _['poolId']
            self.taskId = _['taskId']
            self.status = _['status']
            self.attachment = record['attachment']
            self.attachment_length = len(record['attachment'])
            self.metadata = record.get('metadata', None)
            self.operators = _['taskParams']['operators']
            self.annotations = _['result']['annotations']

    @staticmethod
    def get_annotation_struct(annotation, annotation_type):
        return {
            "key": annotation['key'],
            "label": annotation['label'],
            "type": annotation['type'],
            annotation_type: []
        }

    @staticmethod
    def inter_polate_number(left_index, right_index, target, right, left):
        return left + ((right - left) / (int(right_index) - int(left_index))) * (target - int(left_index))

    @staticmethod
    def angle_closer_to(left_rotation, right_rotation):
        if left_rotation - right_rotation > math.pi:
            return right_rotation + 2 * math.pi
        if left_rotation - right_rotation < -math.pi:
            return right_rotation - 2 * math.pi
        return right_rotation

    @staticmethod
    def get_slot_structure(example):
        for _ in FrameCapture.slot_attr_list:
            if hasattr(example, _) and eval(f'example.{_}'):
                return True
            else:
                continue
        return False

    @staticmethod
    def binary_search(lst, target):
        left = 0
        right = len(lst) - 1

        while left <= right:
            mid = (left + right) // 2
            if lst[mid] == target:
                return mid
            elif lst[mid] < target:
                left = mid + 1
            else:
                right = mid - 1

        return lst[left - 1], lst[left]

    @staticmethod
    def get_world(coordinate):
        ego_heading = coordinate["egoHeading"]
        ego_position = coordinate["ego"]
        ego = Ego(
            heading=(ego_heading['x'], ego_heading['y'], ego_heading['z'], ego_heading['w']),
            position=(ego_position['x'], ego_position['y'], ego_position['z'])
        )
        return ego.ego_rt


class UtilityCategory(object):
    """Shared fields"""

    def __init__(self, data):
        self.label = data['label']
        self.type = data['type']

    @staticmethod
    def get_index(frame_list, count):
        for i in frame_list:
            if count < int(i):
                return frame_list.index(i)

    @staticmethod
    def find_closest_values(numbers, target):
        closest_down = None
        closest_up = None

        for number in numbers:
            if int(number) <= target:
                closest_down = number  # Update closest lower value
            else:
                closest_up = number  # Update closest upper value
                break  # Stop once the upper bound is found

        return closest_down, closest_up



import os
import json
import shutil
import argparse
from tqdm import tqdm
from stardust.convertion import *
from stardust.utils.convert import read_rosetta


def _export(args):
    # Rosetta project ID
    project_id = int(args.project_id)
    # Pool identifiers for the project
    pool_lst = args.pool
    # Export data type
    e_type = args.type
    # Output directory
    save_path = args.output

    # Prepare the output directory
    if os.path.exists(save_path):
        shutil.rmtree(save_path, ignore_errors=True)
    os.makedirs(save_path)
    save_path = os.path.join(save_path, str(project_id))

    # Export Stardust structure
    if e_type == "stardust":
        # Convert Rosetta export to Stardust format
        sd_generator = read_rosetta(project_id=project_id,
                                    pool_lst=pool_lst,
                                    input_path=save_path,
                                    export_type="json",
                                    )
        res_save_path = os.path.join(save_path, f"{project_id}_sd.jsonl")
        with open(res_save_path, 'w', encoding='utf-8') as f:
            for sd_data in tqdm(sd_generator, desc="Processing", unit="item"):
                json.dump(sd_data, f)
                f.write('\n')
        print(f"project id: {project_id}, saved to {res_save_path}")

    # Export COCO structure
    elif e_type == "coco":
        # Convert Rosetta export to Stardust format first
        sd_generator = read_rosetta(project_id=project_id,
                                    pool_lst=pool_lst,
                                    input_path=save_path,
                                    )
        res_save_path = os.path.join(save_path, f"{project_id}_coco.json")
        coco_export(sd_generator, res_save_path)

    # Export KITTI structure
    elif e_type == "kitti":
        # Decide whether frames need to be separated
        _json_data = read_rosetta(project_id=project_id,
                                  pool_lst=pool_lst,
                                  input_path=save_path,
                                  )
        kitti_path = os.path.join(save_path, 'kitti')
        kitti_export(_json_data, kitti_path)

    elif e_type == "rosetta":
        # Decide whether frames need to be separated
        _json_data = read_rosetta(project_id=project_id,
                                  pool_lst=pool_lst,
                                  input_path=save_path,
                                  )
        next(_json_data)


    else:
        print("Unsupported export structure")


def _algo(args):
    print(f'Running algorithm command with args: {args}')


def parse_integer_list(string):
    try:
        # Split comma-separated values and convert to integers
        integer_list = [int(num) for num in string.split(',')]
        return integer_list
    except ValueError:
        # Raise if any value cannot be converted
        raise argparse.ArgumentTypeError("Invalid integer list: {}".format(string))


def main():
    # Create the top-level parser
    parser = argparse.ArgumentParser(description='Stardust command line tool')

    # Define sub-commands
    subparsers = parser.add_subparsers(title='Commands', required=True, dest='command')

    # Export command supporting stardust, coco, kitti, pandaset outputs
    comd_export = subparsers.add_parser('export', help='Export data from Rosetta')
    comd_export.add_argument('--type', "-t", default="stardust", help='Supported export formats: stardust, rosetta, coco, kitti, pandaset')
    comd_export.add_argument('--output', "-o", required=True, help='Output directory for exported data')
    comd_export.add_argument('--project_id', '-p', required=True, help='Rosetta project ID')
    comd_export.add_argument('--pool', '-pool', required=True, help='Comma separated pool IDs for the project', type=parse_integer_list)
    comd_export.set_defaults(func=_export)

    # Placeholder for algorithm-related commands
    comd_algo = subparsers.add_parser('algo', help='Algorithm utilities')
    comd_algo.add_argument('arg1', help='Placeholder parameter')
    comd_algo.set_defaults(func=_algo)

    # Parse CLI arguments
    args = parser.parse_args()

    # Dispatch to the appropriate handler
    args.func(args)


if __name__ == '__main__':
    main()



import numpy as np

from stardust.components.annotations.base_annotation import AnnotationType, AnnotationObject, NumberFiled


class Point(AnnotationObject):

    def __init__(self,
                 x,
                 y,
                 **kwargs) -> None:
        """
        This class defines the concept of 2d geometry point

        Args:
            x: a float number
            y: a float number

        Returns:
            Point instance

        Examples:
            .. code-block:: python

                p = Point(11.12, 345.43)
        """
        self.x = x
        self.y = y
        AnnotationObject.__init__(self, **kwargs)

    def to_array(self) -> np.array:
        return np.array([self.x, self.y])

    @classmethod
    def gen_point(cls, slot, children_lst, parent_id, label_kind=None):
        """
        generate the Polygon obj

        Args:
            slot:
                rosetta slot
            children_lst:
                rosetta children
            parent_id:
                Upper floor ID
        Returns:
            a Polygon instance
        """
        children_lst = children_lst if children_lst else {}
        parent_id = parent_id if parent_id else ""
        label_kind = label_kind if label_kind else ""
        point = Point(
            slot['point']['x'],
            slot['point']['y'],
            parent=parent_id,
            id=slot['id'],
            label=label_kind,
            children=children_lst
        )
        return point



import os
import math
import json
from typing import List, Generator

import numpy as np
from scipy.spatial import ConvexHull

from stardust.components.annotations.box import Box2D
from stardust.components.annotations.box3d import Box3D
from stardust.components.annotations.point import Point
from stardust.components.annotations.polygon import Polygon
from .utils import euclidean_distance, bounding_box, intersection_box, box_corners_bev


class BoxIou2D:
    def __repr__(self):
        description = 'Calculate IoU with 2D boxes'
        return description

    def _intersection_over_union(self, box1: Box2D, box2: Box2D):

        """
        Compute IntersectionOverUnion of box1 and box2

        Args:
            box1: Box2D

            box2: Box2D

        Returns:
            float

        Examples:
            .. code-block:: python

                from stardust.metric.object_detection import BoxIou2D
                from stardust.components.annotations.box import Box2D
                metric = BoxIou2D()
                box1 = Box2D(center=[473.07, 395.93], size=[38.65, 28.67])
                box2 = Box2D(center=[473.07, 395.93], size=[38.65, 28.67])
                IoU = metric._intersection_over_union(box1, box2)
        """
        InterBox = intersection_box(box1, box2)

        IoU = InterBox.area / (box1.area + box2.area - InterBox.area + 1e-6)
        return IoU

    def _generalized_intersection_over_union(self, box1: Box2D, box2: Box2D):
        """
        Compute GeneralizedIntersectionOverUnion of box1 and box2

        Args:
            box1: Box2D

            box2: Box2D

        Returns:
            float

        Examples:
            .. code-block:: python

                from stardust.metric.object_detection import BoxIou2D
                from stardust.components.annotations.box import Box2D
                metric = BoxIou2D()
                box1 = Box2D(center=[473.07, 395.93], size=[38.65, 28.67])
                box2 = Box2D(center=[473.07, 395.93], size=[38.65, 28.67])
                IoU = metric._generalized_intersection_over_union(box1, box2)
        """
        BoundBox = bounding_box(box1, box2)
        InterBox = intersection_box(box1, box2)
        union_area = box1.area + box2.area - InterBox.area
        GIoU = InterBox.area / \
               (union_area - (BoundBox.area - union_area) / BoundBox.area + 1e-6)
        return GIoU

    def _distance_intersection_over_union(self, box1: Box2D, box2: Box2D):
        """
        Compute DistanceIntersectionOverUnion of box1 and box2

        Args:
            box1: Box2D

            box2: Box2D

        Returns:
            float

        Examples:
            .. code-block:: python

                from stardust.metric.object_detection import BoxIou2D
                from stardust.components.annotations.box import Box2D
                metric = BoxIou2D()
                box1 = Box2D(center=[473.07, 395.93], size=[38.65, 28.67])
                box2 = Box2D(center=[473.07, 395.93], size=[38.65, 28.67])
                IoU = metric._distance_intersection_over_union(box1, box2)
        """
        BoundBox = bounding_box(box1, box2)
        BoundDiagonalDistance = euclidean_distance(Point(BoundBox.p1.x, BoundBox.p1.y), Point(
            BoundBox.p2.x, BoundBox.p2.y))
        center_distance = euclidean_distance(box1.center, box2.center)
        DIoU = self._intersection_over_union(
            box1, box2) - (center_distance ** 2) / (BoundDiagonalDistance ** 2)
        return DIoU

    def _complete_intersection_over_union(self, box1: Box2D, box2: Box2D):
        """
        Compute CompleteIntersectionOverUnion of box1 and box2

        Args:
            box1: Box2D

            box2: Box2D

        Returns:
            float

        Examples:
            .. code-block:: python

                from stardust.metric.object_detection import BoxIou2D
                from stardust.components.annotations.box import Box2D
                metric = BoxIou2D()
                box1 = Box2D(center=[473.07, 395.93], size=[38.65, 28.67])
                box2 = Box2D(center=[473.07, 395.93], size=[38.65, 28.67])
                IoU = metric._complete_intersection_over_union(box1, box2)
        """
        v = 4 / (math.pi ** 2) * (math.atan(box2.width / box2.height) -
                                  math.atan(box1.width / box1.height)) ** 2
        IoU = self._intersection_over_union(box1, box2)
        alpha = v / (1 - IoU + v)
        CIoU = self._distance_intersection_over_union(box1, box2) - alpha * v
        return CIoU

    def compute_IoU(self, box1: Box2D, box2: Box2D, IoU_mode: str):

        """
        Computing IoU with the given IoU compute method

        Args:
            box1: Box2D

            box2: Box2D

            IoU_mode: str
                The method to compute IoU, IoU_mode should be chosen from 'IoU', 'GIoU', 'DIoU' and 'CIoU'

        Returns:
            float:
                IoU of box1 and box2

        Examples:
            .. code-block:: python

                from stardust.metric.object_detection import BoxIou2D
                from stardust.components.annotations.box import Box2D
                metric = BoxIou2D()
                box1 = Box2D(center=[473.07, 395.93], size=[38.65, 28.67])
                box2 = Box2D(center=[473.07, 395.93], size=[38.65, 28.67])
                IoU = metric.compute_IoU(box1, box2, 'IoU')
        """
        if IoU_mode not in ['IoU', 'GIoU', 'DIoU', 'CIoU']:
            raise ValueError(
                "Got an invalid IoU_mode, IoU_mode should be in ['IoU', 'GIoU', 'DIoU', 'CIoU']")

        if IoU_mode == 'IoU':
            return self._intersection_over_union(box1, box2)

        elif IoU_mode == 'GIoU':
            return self._generalized_intersection_over_union(box1, box2)

        elif IoU_mode == 'DIoU':
            return self._distance_intersection_over_union(box1, box2)

        elif IoU_mode == 'CIoU':
            return self._complete_intersection_over_union(box1, box2)


class BoxIou3D:
    def __repr__(self):
        description = 'Calculate IoU with 3D boxes'
        return description

    def _intersection_over_union(self, box1: Box3D, box2: Box3D):
        """
        Compute IntersectionOverUnion of box1 and box2

        Args:
            box1: Box3D

            box2: Box3D

        Returns:
            float

        Examples:
            .. code-block:: python

                from stardust.metric.object_detection import BoxIou3D
                from stardust.components.annotations.box3d import Box3D
                metric = BoxIou3D()
                box1 = Box3D(center = [4.13, -3.77, 0.78], size=[1, 5, 1], rotation=[0, 0, -1.57], rotation_order="XYZ")
                box2 = Box3D(center = [4.13, -3.77, 0.78], size=[1, 5, 1], rotation=[0, 0, -1.57], rotation_order="XYZ")
                IoU = metric._intersection_over_union(box1, box2)
        """

        reca, recb = box_corners_bev(box1), box_corners_bev(box2)
        ha, hb = box1.size[2], box2.size[2]
        za, zb = box1.center[2], box2.center[2]
        overlap_height = max(
            0, min((za + ha / 2) - (zb - hb / 2), (zb + hb / 2) - (za - ha / 2)))

        IntersectionArea = intersection_box(reca, recb).area * overlap_height
        UnionArea = box1.size[0] * box1.size[1] * ha + \
                    box2.size[0] * box2.size[1] * hb - IntersectionArea
        return IntersectionArea / UnionArea

    def _generalized_intersection_over_union(self, box1: Box3D, box2: Box3D):

        """
        Compute GeneralizedIntersectionOverUnion of box1 and box2

        Args:
            box1: Box3D

            box2: Box3D


        Returns:
            float

        Examples:
            .. code-block:: python

                from stardust.metric.object_detection import BoxIou3D
                from stardust.components.annotations.box3d import Box3D
                metric = BoxIou3D()
                box1 = Box3D(center = [4.13, -3.77, 0.78], size=[1, 5, 1], rotation=[0, 0, -1.57], rotation_order="XYZ")
                box2 = Box3D(center = [4.13, -3.77, 0.78], size=[1, 5, 1], rotation=[0, 0, -1.57], rotation_order="XYZ")
                IoU = metric._generalized_intersection_over_union(box1, box2)
        """
        boxa_corners, boxb_corners = box_corners_bev(box1), box_corners_bev(box2)
        reca = Box2D(Point(boxa_corners[0]), Point(boxa_corners[1]), Point(
            boxa_corners[2]), Point(boxa_corners[3]))
        recb = Box2D(Point(boxb_corners[0]), Point(boxb_corners[1]), Point(
            boxb_corners[2]), Point(boxb_corners[3]))
        ha, hb = box1.size.z, box2.size.z
        za, zb = box1.center.z, box2.center.z
        overlap_height = max(
            0, min((za + ha / 2) - (zb - hb / 2), (zb + hb / 2) - (za - ha / 2)))
        union_height = max((za + ha / 2) - (zb - hb / 2),
                           (zb + hb / 2) - (za - ha / 2))
        intersection_volume = intersection_box(reca, recb).area * overlap_height
        union_volume = box1.size.x * box1.size.y * ha + \
                       box2.size.x * box2.size.y * hb - intersection_volume
        all_corners = np.vstack((boxa_corners, boxb_corners))
        convexHull_area = ConvexHull(all_corners)
        convex_corners = all_corners[convexHull_area.vertices]
        convex_corners = list(map(Point, convex_corners))
        convex_area = Polygon(convex_corners).area
        convexHull_volume = convex_area * union_height
        return intersection_volume / union_volume - (convexHull_volume - union_volume) / convexHull_volume

    def compute_IoU(self, box1: Box3D, box2: Box3D, IoU_mode):
        """
        Computing IoU with the target IoU mode

        Args:
            box1: Box3D

            box2: Box3D

            IoU_mode: str
                The method to compute IoU, IoU_mode should be chosen from 'IoU' and 'GIoU'

        Returns:
            float:
                IoU of box1 and box2

        Examples:
            .. code-block:: python

                from stardust.metric.object_detection import BoxIou3D
                from stardust.components.annotations.box3e import Box3D
                metric = BoxIou3D()
                box1 = Box3D(center = [4.13, -3.77, 0.78], size=[1, 5, 1], rotation=[0, 0, -1.57], rotation_order="XYZ")
                box2 = Box3D(center = [4.13, -3.77, 0.78], size=[1, 5, 1], rotation=[0, 0, -1.57], rotation_order="XYZ")
                IoU = metric.compute_IoU(box1, box2, 'IoU')
        """
        if IoU_mode not in ['IoU', 'GIoU']:
            raise ValueError(
                "Got an invalid IoU_mode, IoU_mode should be in ['IoU', 'GIoU']")

        if IoU_mode == 'IoU':
            return self._intersection_over_union(box1, box2)

        elif IoU_mode == 'GIoU':
            return self._generalized_intersection_over_union(box1, box2)


IoUMode = {
    '2D': BoxIou2D(),
    '3D': BoxIou3D()}


def compute_metric_single_frame(gt_boxes: List, pd_boxes: List, IoU_thr: float, box_type: str, IoU_mode: str):
    """
    Computing metric of all objects in a single frame

    Args:
        gt_boxes: List
            Box list of ground truth, each box should be a Box-like object(Box2D or Box3d)
        pd_boxes: List
            Box list of predictions, each box should be a Box-like object(Box2D or Box3d)
        IoU_thr: float
            The iou threshold of tp boxes
        box_type: str
            Choose which type of objects to be computed, box_type should be chosen from '2D' and '3D'
        IoU_mode: str
            Choose which IoU compute method to be used
    Returns:
        Tuple: metric of gt, pd, tp, recall, precision and f1
    """
    gt = len(gt_boxes)
    pd = len(pd_boxes)
    tp = 0
    for box_id in list(gt_boxes.keys()):
        if box_id in pd_boxes:
            IoU = IoUMode[box_type].compute_IoU(
                gt_boxes[box_id], pd_boxes[box_id], IoU_mode)

            if IoU >= IoU_thr:
                tp += 1
    recall = float('nan') if gt == 0 else tp / gt
    precision = float('nan') if pd == 0 else tp / pd
    f1 = float('nan') if recall + precision == 0 else 2 * \
                                                      recall * precision / (recall + precision)
    return gt, pd, tp, recall, precision, f1


def compute_metric(data: Generator, IoU_thr: float, IoU_mode: str, save_path: str):
    """
    Computing IoU of all objects in all frames

    Args:
        data (Generator):
            A generator object to get all information from all frames
        IoU_thr: float
            The iou threshold of tp boxes
        IoU_mode: str
            Which IoU compute method to be used
        save_path: str
            Local path to save metric results

    Returns:
        Tuple:
            The metric of dataset which include two dict, the first represents metric of every single frame and the second represents metric of all frames

    Examples:
        .. code-block:: python

            from stardust.metric.object_detection import compute_metric
            from stardust.rosetta.rosetta_data import RosettaData
            project_id = 856
            Data(project_id, 'top', input_path, True).export()
            json_datas = read_rosetta(project_id=project_id,
                                    input_path=input_path,
                                    )
            metric = compute_metric(json_datas, 0.5, 'IoU', 'local/')
    """
    total_gt_2d = total_pd_2d = total_tp_2d = 0
    total_gt_3d = total_pd_3d = total_tp_3d = 0

    metric_output = {}

    for task_id, json_data in enumerate(data):
        gts_2d = json_data.annotation.box2d_lst
        pds_2d = json_data.prediction.box2d_lst

        gts_3d = json_data.annotation.box3d_lst
        pds_3d = json_data.prediction.box3d_lst

        gt_2d, pd_2d, tp_2d, recall_2d, precision_2d, f1_2d = compute_metric_single_frame(
            gts_2d, pds_2d, IoU_thr, '2D', IoU_mode)
        metric_output[task_id] = dict(gt_2d=gt_2d, pd_2d=pd_2d, tp_2d=tp_2d,
                                      recall_2d=recall_2d, precision_2d=precision_2d, f1_2d=f1_2d)
        total_gt_2d += gt_2d
        total_pd_2d += pd_2d
        total_tp_2d += tp_2d

        gt_3d, pd_3d, tp_3d, recall_3d, precision_3d, f1_3d = compute_metric_single_frame(
            gts_3d, pds_3d, IoU_thr, '3D', IoU_mode)
        metric_output[task_id].update(dict(
            gt_3d=gt_3d, pd_3d=pd_3d, tp_3d=tp_3d, recall_3d=recall_3d, precision_3d=precision_3d, f1_3d=f1_3d))
        total_gt_3d += gt_3d
        total_pd_3d += pd_3d
        total_tp_3d += tp_3d

    total_recall_2d = float(
        'nan') if total_gt_2d == 0 else total_tp_2d / total_gt_2d
    total_precision_2d = float(
        'nan') if total_pd_2d == 0 else total_tp_2d / total_pd_2d
    total_f1_2d = float('nan') if total_recall_2d + total_precision_2d == 0 else 2 * \
                                                                                 total_recall_2d * total_precision_2d / \
                                                                                 (total_recall_2d + total_precision_2d)

    total_recall_3d = float(
        'nan') if total_gt_3d == 0 else total_tp_3d / total_gt_3d
    total_precision_3d = float(
        'nan') if total_pd_3d == 0 else total_tp_3d / total_pd_3d
    total_f1_3d = float('nan') if total_recall_3d + total_precision_3d == 0 else 2 * \
                                                                                 total_recall_3d * total_precision_3d / \
                                                                                 (total_recall_3d + total_precision_3d)

    metric_total = dict(gt_2d=total_gt_2d, pd_2d=total_pd_2d, tp_2d=total_tp_2d,
                        recall_2d=total_recall_2d, precision_2d=total_precision_2d, f1_2d=total_f1_2d)
    metric_total.update(dict(gt_3d=total_gt_3d, pd_3d=total_pd_3d, tp_3d=total_tp_3d,
                             recall_3d=total_recall_3d, precision_3d=total_precision_3d, f1_3d=total_f1_3d))
    if save_path:
        os.makedirs(os.path.join(save_path, 'metric', 'object_detection'), exist_ok=True)
        with open(os.path.join(save_path, 'metric', 'object_detection', 'metric_by_task_id.json'), 'w') as f:
            json.dump(metric_output, f)
        with open(os.path.join(save_path, 'metric', 'object_detection', 'metric_summary.json'), 'w') as f:
            json.dump(metric_total, f)

    # return metric_output, metric_total



import numpy as np

from stardust.geometry.ops3d.projection.helper import decode_cam_sources
from stardust.geometry.ops3d.projection.project2img import fisheye_cam2img


def get_img_points(obj_points,
                   num_cameras,
                   img_infos,
                   input_type=None,
                   box_coor=None):
    """

    Args:
        obj_points:
            points of object
        num_cameras:
            number of cameras
        imgInfos: List

        box_coor: input identifier,
            its value is not None when input is box coordinates

    Returns:

    """
    res = {}
    for i in range(num_cameras):
        box = obj_points
        imgs_info = decode_cam_sources(img_infos[i])
        transform_matrix = imgs_info['lidar2cam']
        intrinsic = imgs_info['intrin']
        skew = imgs_info['radial']
        cam_type = imgs_info['cam_type']
        w, h = imgs_info['img_width'], imgs_info['img_height']
        if box_coor != None:
            box = box_coor  # Input already represents 3D box coordinates
        boxes = np.hstack((box, np.ones((box.shape[0], 1))))  # N * 4
        cam_boxes = np.matmul(boxes, transform_matrix.T)
        points_num = cam_boxes.shape[0]
        mask = cam_boxes[:, 2] > 0
        cam_boxes = cam_boxes[mask]
        points_num_after_mask = cam_boxes.shape[0]
        points_occluded = points_num - points_num_after_mask
        if cam_boxes.shape[0] <= 0:
            continue
        if cam_type != 'Fisheye':
            cam_boxes = np.matmul(cam_boxes, intrinsic.T)
            if np.any((np.abs(cam_boxes[:, 2]) < 1)):
                continue
            x, y = cam_boxes[:, 0] / (cam_boxes[:, 2] + 1e-20), cam_boxes[:, 1] / (cam_boxes[:, 2] + 1e-20)
        else:
            cam_in_img = fisheye_cam2img(cam_boxes, skew, 0, intrinsic)
            x = cam_in_img[:, 0]
            y = cam_in_img[:, 1]
        res[i] = (x, y, h, w, points_occluded)
    return res



import math
from typing import List
import numpy as np
from stardust.components.annotations.point import Point
from stardust.components.annotations.point3d import Point3D
from stardust.components.annotations.box import Box2D
from stardust.components.annotations.box3d import Box3D


def euclidean_distance(p1: Point, p2: Point) -> float:
    """
    Compute Euclidean distance of p1 and p2

    Args:
        p1: Point

        p2: Point

    Returns
        float
    """
    assert isinstance(p1, Point) and isinstance(p1, Point)
    x1, y1 = p1.x, p1.y
    x2, y2 = p2.x, p2.y
    return math.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2)


def euclidean_distance_3d(p1: Point3D, p2: Point3D) -> float:
    """
    Compute Euclidean distance of p1 and p2

    Args:
        p1: Point3D

        p2: Point3D

    Returns
        float
    """
    assert isinstance(p1, Point3D) and isinstance(p1, Point3D)
    x1, y1, z1 = p1.x, p1.y, p1.z
    x2, y2, z2 = p2.x, p2.y, p2.z
    return math.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2 + (z2 - z1) ** 2)


def bounding_box(box1: Box2D, box2: Box2D) -> Box2D:
    """
    Compute bounding_box of box1 and box2

    Args:
        box1: Box2D

        box2: Box2D

    Returns
        Box2D
    """
    x11, y11, x12, y12 = box1.p1[0], box1.p1[1], box1.p2[0], box1.p2[1]
    x21, y21, x22, y22 = box2.p1[0], box2.p1[1], box2.p2[0], box2.p2[1]
    x1 = min(x11, x21)
    y1 = min(y11, y21)
    x2 = max(x12, x22)
    y2 = max(y12, y22)
    return Box2D(Point(x1, y1), Point(x2, y2))


def intersection_box(box1: Box2D, box2: Box2D) -> Box2D:
    """
    Compute intersection box of box1 and box2

    Args:
        box1: Box2D

        box2: Box2D

    Returns:
        Box2D
    """
    x11, y11, x12, y12 = box1.p1[0], box1.p1[1], box1.p2[0], box1.p2[1]
    x21, y21, x22, y22 = box2.p1[0], box2.p1[1], box2.p2[0], box2.p2[1]
    x1, y1 = max(x11, x21), max(y11, y21)
    x2, y2 = min(x12, x22), min(y12, y22)
    if x1 >= x2 or y1 >= y2:
        return Box2D(Point(0, 0), Point(0, 0))
    return Box2D(Point(x1, y1), Point(x2, y2))


def box_corners_bev(box: Box3D):
    """
    Get box corners of box in bev

    Args:
        box: Box3D

    Returns
        Box2D
    """
    x, y, l, w, r = box.center[0], box.center[1], box.size[0], box.size[1], box.rotation[2]
    BottomPlaneCenter = np.array([x, y])
    cos, sin = np.cos(r), np.sin(r)
    pc0 = np.array([x + cos * l / 2 + sin * w / 2,
                    y + sin * l / 2 - cos * w / 2])
    pc1 = 2 * BottomPlaneCenter - pc0
    return Box2D(p1=Point(pc1[0], pc1[1]), p2=Point(pc0[0], pc0[1]))



import numpy as np

from stardust.utils.frames import *
from stardust.utils.matrix4 import *


class Slot:
    attr_map = {
        'input': 'input',
        'point': 'point',
        'line': 'vertices',
        'splines': 'vertices',
        'polygon': 'vertices',
        'box2d': 'plane',
        'cuboid': 'vertices',
        'point3d': 'point',
        'line3d': 'points',
        'polygon3d': 'points',
        'box3d': 'matrix',

        'point3d_world': 'point',
        'line3d_world': 'points',
        'polygon3d_world': 'points',
        'box3d_world': 'matrix',
    }

    @classmethod
    def preposition(cls, data, frame_number, attachment):
        cls.target = frame_number
        cls.attachment = attachment
        if data.get('input'):
            cls.data = data['input']
        else:
            cls.data = data
        cls.time_series = cls.data.get('timeSeries', {})
        if cls.time_series:
            cls.frame_list = np.array(list(cls.time_series.keys()), dtype=int).tolist()

    def __init__(self, data, frame_number, attachment=None):
        self.preposition(data, frame_number, attachment)
        for n in ['id', 'label', 'teamId',  'source', 'coordinate']:
            if data.get(n):
                setattr(self, n, data[n])
        self.type = data['type']
        if data.get('coordinate', '') == 'world':
            self.type = f'{self.type}_world'
        commonality_result = self.commonality(self.type)
        if callable(commonality_result):
            if isinstance(commonality_result()(), bool):
                self.label = True
            setattr(self, self.attr_map[self.type], commonality_result()())
        else:
            if isinstance(commonality_result, bool):
                self.label = True
            setattr(self, self.attr_map[self.type], commonality_result)
        self.type = data['type']

    def commonality(self, key):
        if self.time_series:
            temp = str(self.frame_list[-1])
            # Interpolate toward earlier frames
            if self.target < self.frame_list[0]:
                return True
            # Interpolate forward when last state is "appear"
            if self.time_series[temp]['type'] == 'appear' and self.target > self.frame_list[-1]:
                if key == 'box3d':
                    setattr(self, 'box', self.time_series[temp]['value'])
                    return Matrix4(tuple(self.time_series[temp]['value'])).box_info
                elif key == 'input':
                    return {'value': self.time_series[temp]['value']}
                elif key == 'point3d_world':
                    return self.point3d_one_now([self.time_series[temp]['value']], self.frame_list[-1])[0]
                elif key in ['line3d_world', 'polygon3d_world']:
                    return self.point3d_one_now(self.time_series[temp]['value'], self.frame_list[-1])
                elif key == 'box3d_world':
                    box3d_world = self.box3d_one_now(self.time_series[temp]['value'], self.frame_list[-1])
                    setattr(self, 'box', box3d_world)
                    return Matrix4(tuple(box3d_world)).box_info
                return self.time_series[temp]['value']
            # Interpolate forward when last state is "disappear"
            elif self.time_series[temp]['type'] == 'disappear' and self.target > self.frame_list[-1]:
                return True
            else:
                # Return value directly if frame exists
                if self.target in self.frame_list:
                    if self.time_series[str(self.target)]['type'] == "disappear":
                        return True
                    input_value = self.time_series[str(self.target)].get("value")
                    if key in ['box3d', 'teamId', 'box3d_world']:
                        setattr(self, 'box', input_value)
                        return Matrix4(tuple(input_value)).box_info
                    elif key == 'input':
                        return {'value': input_value,
                                'teamId': self.time_series[str(self.target)].get("teamId")
                                }
                    return input_value
                else:
                    return self.slot_factory
        else:
            return True

    def slot_factory(self):
        return {
            'input': self.input_time_series,
            'point': self.point_time_series,
            'line': self.line_time_series,
            'splines': self.line_time_series,
            'polygon': self.line_time_series,
            'box2d': self.box2d_time_series,
            'cuboid': self.cuboid_time_series,
            'point3d': self.point3d_time_series,
            'line3d': self.line3d_time_series,
            'polygon3d': self.line3d_time_series,
            'box3d': self.box3d_time_series,
            'point3d_world': self.point3d_world_time_series,
            'line3d_world': self.line3d_world_time_series,
            'polygon3d_world': self.line3d_world_time_series,
            'box3d_world': self.box3d_world_time_series,

        }[self.type]

    @staticmethod
    def _lerp_factor(left_index: int, right_index: int, target: int) -> float:
        denominator = right_index - left_index
        if denominator == 0:
            return 0.0
        return (target - left_index) / denominator

    @staticmethod
    def _lerp_dict(left_dict: dict, right_dict: dict, keys: tuple, factor: float) -> dict:
        left_vec = np.array([left_dict[key] for key in keys], dtype=float)
        right_vec = np.array([right_dict[key] for key in keys], dtype=float)
        blended = left_vec + (right_vec - left_vec) * factor
        return {key: float(value) for key, value in zip(keys, blended)}

    @staticmethod
    def _lerp_point_list(left_points: list, right_points: list, keys: tuple, factor: float) -> list:
        left_arr = np.array([[point[key] for key in keys] for point in left_points], dtype=float)
        right_arr = np.array([[point[key] for key in keys] for point in right_points], dtype=float)
        blended = left_arr + (right_arr - left_arr) * factor
        return [
            {key: float(value) for key, value in zip(keys, row)}
            for row in blended
        ]

    def input_time_series(self):
        left, right = binary_search(self.frame_list, self.target)
        if self.time_series[str(left)]['type'] == 'appear':
            return {'value': self.time_series[str(left)]['value']}
        return True

    def l_or_r(self, left: int, right: int):
        return (left, right, self.target), self.time_series[str(left)]['value'], self.time_series[str(right)]['value']

    # 2D point
    def point_time_series(self):
        left, right = binary_search(self.frame_list, self.target)
        if self.time_series[str(left)]['type'] == 'appear':
            if self.time_series[str(right)]['type'] == 'appear':
                temp, l, r = self.l_or_r(left, right)
                factor = self._lerp_factor(*temp)
                return self._lerp_dict(l, r, ('x', 'y'), factor)
            return self.time_series[str(left)]['value']
        return True

    # 2D line, curve, or polygon
    def line_time_series(self):
        left, right = binary_search(self.frame_list, self.target)
        if self.time_series[str(left)]['type'] == 'appear':
            if self.time_series[str(right)]['type'] == 'appear':
                temp, l, r = self.l_or_r(left, right)
                if len(l) == len(r):
                    factor = self._lerp_factor(*temp)
                    return self._lerp_point_list(l, r, ('x', 'y'), factor)
                return l
            return self.time_series[str(left)]['value']
        return True

    # 2D bounding box
    def box2d_time_series(self):
        left, right = binary_search(self.frame_list, self.target)
        if self.time_series[str(left)]['type'] == 'appear':
            if self.time_series[str(right)]['type'] == 'appear':
                temp, l, r = self.l_or_r(left, right)
                factor = self._lerp_factor(*temp)
                corners = ('topLeft', 'topRight', 'bottomLeft', 'bottomRight')
                return {
                    corner: self._lerp_dict(l[corner], r[corner], ('x', 'y'), factor)
                    for corner in corners
                }
            return self.time_series[str(left)]['value']
        return True

    # 2D projection of cuboid
    def cuboid_time_series(self):
        left, right = binary_search(self.frame_list, self.target)
        if self.time_series[str(left)]['type'] == 'appear':
            if self.time_series[str(right)]['type'] == 'appear':
                temp, l, r = self.l_or_r(left, right)
                factor = self._lerp_factor(*temp)
                faces = ('front', 'back')
                corners = ('topLeft', 'topRight', 'bottomLeft', 'bottomRight')
                return {
                    face: {
                        corner: self._lerp_dict(l[face][corner], r[face][corner], ('x', 'y'), factor)
                        for corner in corners
                    }
                    for face in faces
                }
            return self.time_series[str(left)]['value']
        return True

    # 3D point
    def point3d_time_series(self):
        left, right = binary_search(self.frame_list, self.target)
        if self.time_series[str(left)]['type'] == 'appear':
            if self.time_series[str(right)]['type'] == 'appear':
                temp, l, r = self.l_or_r(left, right)
                factor = self._lerp_factor(*temp)
                return self._lerp_dict(l, r, ('x', 'y', 'z'), factor)
            return self.time_series[str(left)]['value']
        return True

    # 3D line or polygon
    def line3d_time_series(self):
        left, right = binary_search(self.frame_list, self.target)
        if self.time_series[str(left)]['type'] == 'appear':
            if self.time_series[str(right)]['type'] == 'appear':
                temp, l, r = self.l_or_r(left, right)
                if len(l) == len(r):
                    factor = self._lerp_factor(*temp)
                    return self._lerp_point_list(l, r, ('x', 'y', 'z'), factor)
                return l
            return self.time_series[str(left)]['value']
        return True

    @staticmethod
    def matrix(temp: tuple, left: dict, right: dict):
        return {
            'position': {
                'x': inter_polate_number(*temp, left['position']['x'], right['position']['x']),
                'y': inter_polate_number(*temp, left['position']['y'], right['position']['y']),
                'z': inter_polate_number(*temp, left['position']['z'], right['position']['z'])
            },
            'scale': {
                'length': inter_polate_number(*temp, left['scale']['length'], right['scale']['length']),
                'width': inter_polate_number(*temp, left['scale']['width'], right['scale']['width']),
                'height': inter_polate_number(*temp, left['scale']['height'], right['scale']['height'])
            },
            'rotation': {
                'x': inter_polate_number(*temp, left['rotation']['x'],
                                         angle_closer_to(left['rotation']['x'], right['rotation']['x'])),
                'y': inter_polate_number(*temp, left['rotation']['y'],
                                         angle_closer_to(left['rotation']['y'], right['rotation']['y'])),
                'z': inter_polate_number(*temp, left['rotation']['z'],
                                         angle_closer_to(left['rotation']['z'], right['rotation']['z'])),
                '_order': 'XYZ'
            }
        }

    # 3D bounding box
    def box3d_time_series(self):
        left, right = binary_search(self.frame_list, self.target)
        if self.time_series[str(left)]['type'] == 'appear':
            if self.time_series[str(right)]['type'] == 'appear':
                temp, l, r = self.l_or_r(left, right)
                l_box, r_box = Matrix4(tuple(l)).box_info, Matrix4(tuple(r)).box_info
                matrix = self.matrix(temp, l_box, r_box)
                setattr(self, 'box', Matrix4.get_transform_matrix(matrix))
                return matrix
            setattr(self, 'box', self.time_series[str(left)]['value'])
            return Matrix4(tuple(self.time_series[str(left)]['value'])).box_info
        return True

    def point3d_one_now(self, value: list, frame: int):
        value = point3d_to_world(value, self.attachment[frame].get('coordinate', []))
        return point3d_to_world(value, self.attachment[self.target].get('coordinate', []), overturn=True)

    def point3d_world(self, left: int, right: int):
        temp, l, r = self.l_or_r(left, right)
        l_ = point3d_to_world(l, self.attachment[left].get('coordinate', []))
        r_ = point3d_to_world(r, self.attachment[right].get('coordinate', []))
        return temp, l_, r_

    def point3d_now(self, temp: tuple, left: dict, right: dict):
        point3d_world = [[inter_polate_number(*temp, left[n][i], right[n][i]) for i in ['x', 'y', 'z']]
                         for n in range(len(right))]
        return point3d_to_world(point3d_world, self.attachment[self.target].get('coordinate', []), overturn=True)

    # 3D point (world coordinates)
    def point3d_world_time_series(self):
        left, right = binary_search(self.frame_list, self.target)
        if self.time_series[str(left)]['type'] == 'appear':
            if self.time_series[str(right)]['type'] == 'appear':
                temp, l_world, r_world = self.point3d_world(left, right)
                return self.point3d_now(temp, l_world, r_world)[0]
            return self.point3d_one_now([self.time_series[left]['value']], self.target)[0]
        return True

    # 3D line or polygon (world coordinates)
    def line3d_world_time_series(self):
        left, right = binary_search(self.frame_list, self.target)
        if self.time_series[str(left)]['type'] == 'appear':
            if self.time_series[str(right)]['type'] == 'appear':
                temp, l_world, r_world = self.point3d_world(left, right)
                if len(l_world) == len(r_world):
                    return self.point3d_now(temp, l_world, r_world)
                point3d_world = [[l_world[n][i] for i in ['x', 'y', 'z']] for n in range(len(l_world))]
                return point3d_to_world(point3d_world, self.attachment[self.target].get('coordinate', []),
                                        overturn=True)
            return self.point3d_one_now([self.time_series[left]['value']], self.target)
        return True

    def box3d_one_now(self, value: list, frame: int):
        value = box3d_to_world(value, self.attachment[frame].get('coordinate', []))
        return box3d_to_world(value, self.attachment[self.target].get('coordinate', []), overturn=True)

    def box3d_world(self, left: int, right: int):
        temp, l, r = self.l_or_r(left, right)
        l_ = box3d_to_world(l, self.attachment[left].get('coordinate', []))
        r_ = box3d_to_world(r, self.attachment[right].get('coordinate', []))
        return temp, l_, r_

    # 3D bounding box (world coordinates)
    def box3d_world_time_series(self):
        left, right = binary_search(self.frame_list, self.target)
        if self.time_series[str(left)]['type'] == 'appear':
            if self.time_series[str(right)]['type'] == 'appear':
                temp, l_world, r_world = self.box3d_world(left, right)
                l_world_box, r_world_box = Matrix4(tuple(l_world)).box_info, Matrix4(tuple(r_world)).box_info
                matrix = self.matrix(temp, l_world_box, r_world_box)
                box = box3d_to_world(Matrix4.get_transform_matrix(matrix),
                                     self.attachment[self.target].get('coordinate', []), overturn=True)
                setattr(self, 'box', box)
                return Matrix4(tuple(box)).box_info
            box3d_world = self.box3d_one_now(self.time_series[left]['value'], self.target)
            setattr(self, 'box', box3d_world)
            return Matrix4(tuple(box3d_world)).box_info
        return True


if __name__ == '__main__':
    pass



import math
import numpy as np
from typing import Optional, Union, Tuple
from scipy.spatial.transform import Rotation


class Ego:
    def __init__(self,
                 heading: Optional[tuple] = None,
                 position: Optional[tuple] = None,
                 longitude: Optional[Union[float, int]] = None,
                 latitude: Optional[Union[float, int]] = None,
                 altitude: Optional[Union[float, int]] = None
                 ):
        """
        Ego

        Args:
            heading:
                quaternion (x, y, z, w)
            position:
                offset (x, y, z)
            longitude:
                the longitude
            latitude:
                the latitude
            altitude:
                the altitude

        Returns:
            Ego instance

        """
        self.heading = heading
        self.position = position
        self.longitude = longitude
        self.latitude = latitude
        self.altitude = altitude

    def __repr__(self):
        return "the ego param"

    @property
    def ego_rt(self) -> np.matrix:
        """
        ego rt matrix
        """
        r = Rotation.from_quat(self.heading).as_matrix()
        rt = np.r_[np.c_[r, np.array(self.position)], np.array([[0, 0, 0, 1]])]
        return np.matrix(rt)

    @property
    def rotate(self) -> np.ndarray:
        """
        ego r matrix
        """
        r = Rotation.from_quat(self.heading).as_matrix()
        return r

    @property
    def to_world_coordinates(self) -> Tuple[float, float, float]:
        """
        Longitude and latitude to the coordinates of the world coordinate system
        """
        wgs84_a = 6378137.0  # major axis of the Earth's ellipsoid, in meters
        wgs84_f = 1 / 298.257223563  # flattening of the Earth's ellipsoid
        wgs84_e2 = wgs84_f * (2 - wgs84_f)  # eccentricity squared of the Earth's ellipsoid
        cos_lat = math.cos(self.latitude * math.pi / 180)
        sin_lat = math.sin(self.latitude * math.pi / 180)
        cos_lon = math.cos(self.longitude * math.pi / 180)
        sin_lon = math.sin(self.longitude * math.pi / 180)
        n = wgs84_a / math.sqrt(1 - wgs84_e2 * sin_lat * sin_lat)
        x = (n + self.altitude) * cos_lat * cos_lon
        y = (n + self.altitude) * cos_lat * sin_lon
        z = (n * (1 - wgs84_e2) + self.altitude) * sin_lat
        return x, y, z



from copy import deepcopy
import numpy as np

from stardust.geometry.ops3d.bbox.structure.bbox_np_ops import rbbox3d_to_bev_corners

"""
    Implementation of corner align from paper https://arxiv.org/abs/2101.06586
"""


def rotation_matrix(heading):
    rm = np.array([[np.cos(heading), -np.sin(heading)],
                   [np.sin(heading), np.cos(heading)]])
    return rm


def corner_align(box_3d: np.ndarray, align_size: np.ndarray):
    """
    Implementation of corner align from paper https://arxiv.org/abs/2101.06586
    Align shape based on original box: box_3d and new size: align_size

    Args:
        box_3d:
            box in stardust coordiante, (x, y, z, ..., heading)
        align_size:
            new size, (length on x axis, length on y axis, ..)

    Return:
        Aligned bounding box
    """
    box3d_x, box3d_y, heading = box_3d[0], box_3d[1], box_3d[-1]
    kitti_box3d = deepcopy(box_3d)
    kitti_box3d[-1] = np.pi - kitti_box3d[-1]
    rm_heading = rotation_matrix(kitti_box3d[-1])
    bev_corners = rbbox3d_to_bev_corners(kitti_box3d[np.newaxis, ...])[0]
    align_w, align_l, _ = align_size  # length on x axis, length on y axis
    corner_d = np.linalg.norm(bev_corners, axis=1)
    closest_id = np.argmin(corner_d)
    # transform bev box center to origin
    bev_corners_origin = bev_corners - np.array([box3d_x, box3d_y])
    # rotate
    bev_corners_origin = (bev_corners_origin @ rm_heading.T).astype(np.float32)  # inverse of rm(heading)
    fixed_corner = bev_corners_origin[closest_id]
    fixed_x, fixed_y = fixed_corner
    # loop through the rest three corners
    center_x, center_y = 0, 0
    for i in range(len(bev_corners_origin)):
        if i != closest_id:
            x, y = bev_corners_origin[i]
            if abs(x - fixed_x) < 1e-3:
                if y > fixed_y:
                    center_y = fixed_y + (align_l / 2)
                else:
                    center_y = fixed_y - (align_l / 2)
            if abs(y - fixed_y) < 1e-3:
                if x > fixed_x:
                    center_x = fixed_x + (align_w / 2)
                else:
                    center_x = fixed_x - (align_w / 2)
    new_center = np.array([center_x, center_y])
    new_center = (np.linalg.inv(rm_heading) @ new_center.T)[[1, 0]]
    new_center += np.array([box3d_x, box3d_y])
    box_3d[:2] = new_center
    box_3d[3:6] = align_size
    return box_3d


if __name__ == '__main__':
    box = np.array([138.49, 24.68, 1.88, 9.86, 3.21, 3.56, -1.0692230703881558])
    align_size = np.array([12.73, 3.62, 3.77])
    aligned_box = corner_align(box, align_size)
    print(aligned_box)



import numpy as np
import numba


def rotation_3d_in_axis(points, angles, axis=0):
    rot_sin = np.sin(angles)
    rot_cos = np.cos(angles)
    ones = np.ones_like(rot_cos)
    zeros = np.zeros_like(rot_cos)
    if axis == 1:
        rot_mat_T = np.stack(
            [
                [rot_cos, zeros, -rot_sin],
                [zeros, ones, zeros],
                [rot_sin, zeros, rot_cos],
            ]
        )
    elif axis == 2 or axis == -1:
        rot_mat_T = np.stack(
            [
                [rot_cos, -rot_sin, zeros],
                [rot_sin, rot_cos, zeros],
                [zeros, zeros, ones],
            ]
        )
    elif axis == 0:
        rot_mat_T = np.stack(
            [
                [zeros, rot_cos, -rot_sin],
                [zeros, rot_sin, rot_cos],
                [ones, zeros, zeros],
            ]
        )
    else:
        raise ValueError("axis should in range")

    return np.einsum("aij,jka->aik", points, rot_mat_T)


def center_to_corner_box3d(centers, dims, angles=None, origin=(0.5, 0.5, 0.5), axis=2):
    """
    convert kitti locations, dimensions and angles to corners

    Args:
        centers (float array, shape=[N, 3]): locations in kitti label file.
        dims (float array, shape=[N, 3]): dimensions in kitti label file.
        angles (float array, shape=[N]): rotation_y in kitti label file.
        origin (list or array or float): origin point relate to smallest point.
            use [0.5, 1.0, 0.5] in camera and [0.5, 0.5, 0] in lidar.
        axis (int): rotation axis. 1 for camera and 2 for lidar.

    Returns:
        [type]: [description]
    """
    # 'length' in kitti format is in x axis.
    # yzx(hwl)(kitti label file)<->xyz(lhw)(camera)<->z(-x)(-y)(wlh)(lidar)
    # center in kitti format is [0.5, 1.0, 0.5] in xyz.
    corners = corners_nd(dims, origin=origin)
    if angles is not None:
        corners = rotation_3d_in_axis(corners, angles, axis=axis)
    corners += centers.reshape([-1, 1, 3])
    return corners


def corners_nd(dims, origin=0.5):
    """
    generate relative box corners based on length per dim and origin point.

    Args:
        dims (float array, shape=[N, ndim]): array of length per dim
        origin (list or array or float): origin point relate to smallest point.

    Returns:
        float array, shape=[N, 2 ** ndim, ndim]: returned corners.
        point layout example: (2d) x0y0, x0y1, x1y0, x1y1;
            (3d) x0y0z0, x0y0z1, x0y1z0, x0y1z1, x1y0z0, x1y0z1, x1y1z0, x1y1z1
            where x0 < x1, y0 < y1, z0 < z1
    """
    ndim = int(dims.shape[1])
    corners_norm = np.stack(
        np.unravel_index(np.arange(2 ** ndim), [2] * ndim), axis=1
    ).astype(dims.dtype)
    # now corners_norm has format: (2d) x0y0, x0y1, x1y0, x1y1
    # (3d) x0y0z0, x0y0z1, x0y1z0, x0y1z1, x1y0z0, x1y0z1, x1y1z0, x1y1z1
    # so need to convert to a format which is convenient to do other computing.
    # for 2d boxes, format is clockwise start with minimum point
    # for 3d boxes, please draw lines by your hand.
    if ndim == 2:
        # generate clockwise box corners
        corners_norm = corners_norm[[0, 1, 3, 2]]
    elif ndim == 3:
        corners_norm = corners_norm[[0, 1, 3, 2, 4, 5, 7, 6]]
    corners_norm = corners_norm - np.array(origin, dtype=dims.dtype)
    corners = dims.reshape([-1, 1, ndim]) * corners_norm.reshape([1, 2 ** ndim, ndim])
    return corners


def corner_to_surfaces_3d(corners):
    """
    convert 3d box corners from corner function above to surfaces that normal vectors all direct to internal.

    Args:
        corners (float array, [N, 8, 3]): 3d box corners.

    Returns:
        surfaces (float array, [N, 6, 4, 3]):
    """
    # box_corners: [N, 8, 3], must from corner functions in this module
    surfaces = np.array(
        [
            [corners[:, 0], corners[:, 1], corners[:, 2], corners[:, 3]],
            [corners[:, 7], corners[:, 6], corners[:, 5], corners[:, 4]],
            [corners[:, 0], corners[:, 3], corners[:, 7], corners[:, 4]],
            [corners[:, 1], corners[:, 5], corners[:, 6], corners[:, 2]],
            [corners[:, 0], corners[:, 4], corners[:, 5], corners[:, 1]],
            [corners[:, 3], corners[:, 2], corners[:, 6], corners[:, 7]],
        ]
    ).transpose([2, 0, 1, 3])
    return surfaces


@numba.njit
def surface_equ_3d_jitv2(surfaces):
    # polygon_surfaces: [num_polygon, num_surfaces, num_points_of_polygon, 3]
    num_polygon = surfaces.shape[0]
    max_num_surfaces = surfaces.shape[1]
    normal_vec = np.zeros((num_polygon, max_num_surfaces, 3), dtype=surfaces.dtype)
    d = np.zeros((num_polygon, max_num_surfaces), dtype=surfaces.dtype)
    sv0 = surfaces[0, 0, 0] - surfaces[0, 0, 1]
    sv1 = surfaces[0, 0, 0] - surfaces[0, 0, 1]
    for i in range(num_polygon):
        for j in range(max_num_surfaces):
            sv0[0] = surfaces[i, j, 0, 0] - surfaces[i, j, 1, 0]
            sv0[1] = surfaces[i, j, 0, 1] - surfaces[i, j, 1, 1]
            sv0[2] = surfaces[i, j, 0, 2] - surfaces[i, j, 1, 2]
            sv1[0] = surfaces[i, j, 1, 0] - surfaces[i, j, 2, 0]
            sv1[1] = surfaces[i, j, 1, 1] - surfaces[i, j, 2, 1]
            sv1[2] = surfaces[i, j, 1, 2] - surfaces[i, j, 2, 2]
            normal_vec[i, j, 0] = sv0[1] * sv1[2] - sv0[2] * sv1[1]
            normal_vec[i, j, 1] = sv0[2] * sv1[0] - sv0[0] * sv1[2]
            normal_vec[i, j, 2] = sv0[0] * sv1[1] - sv0[1] * sv1[0]

            d[i, j] = (
                    -surfaces[i, j, 0, 0] * normal_vec[i, j, 0]
                    - surfaces[i, j, 0, 1] * normal_vec[i, j, 1]
                    - surfaces[i, j, 0, 2] * normal_vec[i, j, 2]
            )
    return normal_vec, d


@numba.njit
def _points_in_convex_polygon_3d_jit(
        points, polygon_surfaces, normal_vec, d, num_surfaces=None
):
    """
    check points is in 3d convex polygons.

    Args:
        points: [num_points, 3] array.

        polygon_surfaces: [num_polygon, max_num_surfaces, max_num_points_of_surface, 3]  array.
            all surfaces' normal vector must direct to internal. max_num_points_of_surface must at least 3.
        num_surfaces: [num_polygon] array. indicate how many surfaces
            a polygon contain

    Returns:
        [num_points, num_polygon] bool array.
    """
    max_num_surfaces, max_num_points_of_surface = polygon_surfaces.shape[1:3]
    num_points = points.shape[0]
    num_polygons = polygon_surfaces.shape[0]
    ret = np.ones((num_points, num_polygons), dtype=np.bool_)
    sign = 0.0
    for i in range(num_points):
        for j in range(num_polygons):
            for k in range(max_num_surfaces):
                if k > num_surfaces[j]:
                    break
                sign = (
                        points[i, 0] * normal_vec[j, k, 0]
                        + points[i, 1] * normal_vec[j, k, 1]
                        + points[i, 2] * normal_vec[j, k, 2]
                        + d[j, k]
                )
                if sign >= 0:
                    ret[i, j] = False
                    break
    return ret


def points_in_convex_polygon_3d_jit(points, polygon_surfaces, num_surfaces=None):
    """
    check points is in 3d convex polygons.

    Args:
        points: [num_points, 3] array.

        polygon_surfaces: [num_polygon, max_num_surfaces, max_num_points_of_surface, 3] array.
            all surfaces' normal vector must direct to internal.
            max_num_points_of_surface must at least 3.
        num_surfaces: [num_polygon] array. indicate how many surfaces
            a polygon contain

    Returns:
        [num_points, num_polygon] bool array.
    """
    max_num_surfaces, max_num_points_of_surface = polygon_surfaces.shape[1:3]
    num_points = points.shape[0]
    num_polygons = polygon_surfaces.shape[0]
    if num_surfaces is None:
        num_surfaces = np.full((num_polygons,), 9999999, dtype=np.int64)
    normal_vec, d = surface_equ_3d_jitv2(polygon_surfaces[:, :, :3, :])
    # normal_vec: [num_polygon, max_num_surfaces, 3]
    # d: [num_polygon, max_num_surfaces]
    return _points_in_convex_polygon_3d_jit(
        points, polygon_surfaces, normal_vec, d, num_surfaces
    )


def points_in_rbbox(points, rbbox, z_axis=2, origin=(0.5, 0.5, 0.5)):
    rbbox_corners = center_to_corner_box3d(
        rbbox[:, :3], rbbox[:, 3:6], rbbox[:, -1], origin=origin, axis=z_axis
    )
    surfaces = corner_to_surfaces_3d(rbbox_corners)
    indices = points_in_convex_polygon_3d_jit(points[:, :3], surfaces)
    return indices


def rotation_2d(points, angles):
    """
    rotation 2d points based on origin point clockwise when angle positive.

    Args:
        points (float array, shape=[N, point_size, 2]): points to be rotated.
        angles (float array, shape=[N]): rotation angle.

    Returns:
        float array: same shape as points
    """
    rot_sin = np.sin(angles)
    rot_cos = np.cos(angles)
    rot_mat_T = np.stack([[rot_cos, -rot_sin], [rot_sin, rot_cos]])
    return np.einsum("aij,jka->aik", points, rot_mat_T)


def center_to_corner_box2d(centers, dims, angles=None, origin=0.5):
    """
    convert kitti locations, dimensions and angles to corners.
    format: center(xy), dims(xy), angles(clockwise when positive)

    Args:
        centers (float array, shape=[N, 2]):
            locations in kitti label file.
        dims (float array, shape=[N, 2]):
            dimensions in kitti label file.
        angles (float array, shape=[N]):
            rotation_y in kitti label file.

    Returns:
        [type]: [description]
    """
    # 'length' in kitti format is in x axis.
    # xyz(hwl)(kitti label file)<->xyz(lhw)(camera)<->z(-x)(-y)(wlh)(lidar)
    # center in kitti format is [0.5, 1.0, 0.5] in xyz.
    corners = corners_nd(dims, origin=origin)
    # corners: [N, 4, 2]
    if angles is not None:
        corners = rotation_2d(corners, angles)
    corners += centers.reshape([-1, 1, 2])
    return corners


def rbbox3d_to_bev_corners(rbboxes, origin=0.5):
    return center_to_corner_box2d(
        rbboxes[..., :2], rbboxes[..., 3:5], rbboxes[..., 6], origin
    )



from enum import Enum
from typing import Optional
from scipy.spatial.transform import Rotation
import numpy as np
from typing import Dict
import math
import open3d as o3d


class CameraType(Enum):
    PinHole = "PinHole"
    Fisheye = "Fisheye"
    OmniDirectional = "OmniDirectional"

    @classmethod
    def camera_type_list(cls) -> list:
        return [_.value for _ in CameraType]


class Camera:
    def __init__(self,
                 camera_type: str,
                 heading: Optional[tuple] = None,
                 position: Optional[tuple] = None,
                 intrinsic: Optional[tuple] = None,
                 radial: Optional[tuple] = None,
                 tangential: Optional[tuple] = None,
                 fov: Optional[int] = None,
                 skew: Optional[int] = 0,
                 projection: Optional[tuple] = None,
                 name: str = None
                 ):
        """
        camera object

        Args:
            camera_type:
                Pinholes, fish eyes, panoramic views
            heading:
                quaternion (x, y, z, w)
            position:
                offset (x, y, z)
            intrinsic:
                camera parameter (fx, fy, cx, cy) or (fx, fy, c, d, e, cx, cy)
            radial:
                radial distortion (k1...)
            tangential:
                tangential distortion (p1...)
            fov:
                camera range (angle system)
            projection:
                panoramic camera parameters (p0...)

        Returns:
            Camera inatance

        Examples:
            .. code-block:: python

                cam = Camera(
                    'PinHole',
                    heading=(0.4351436321314605, -0.4677553765588058, 0.5492505771794342, -0.5386824023335351),
                    position=(1.5601294, -0.37986112, 1.3333596),
                    intrinsic=(1846.008599353862, 1857.930995475229, 1948.09353722228, 1078.768188621484),
                    radial=(0, 0, 0, 0),
                    tangential=(0, 0)
                )

        """
        camera_type_list = CameraType.camera_type_list()
        assert camera_type in camera_type_list, f'camera_type is one of "{" ".join(camera_type_list)}"'
        self.camera_type = camera_type
        self.heading = heading
        self.position = position
        self.intrinsic = intrinsic
        self.name = name
        if camera_type == CameraType.OmniDirectional.value:
            assert len(self.intrinsic) == 7, 'intrinsic is fx, fy, c, d, e, cx, cy'
        else:
            assert len(self.intrinsic) == 4, 'intrinsic is fx, fy, cx, cy'
        self.radial = radial
        self.tangential = tangential
        self.fov = fov
        self.projection = projection
        self.skew = skew

    def __repr__(self):
        return "the camera param"

    @property
    def rotated_vector(self):
        """
            Extract the x, y, and z components of the rotated vector
        """
        additional_rotation = R.from_quat(self.heading)
        # Create a vector initially aligned with the x-axis
        vector = np.array([1.0, 0.0, 0.0])
        rotated_vector = additional_rotation.apply(vector)
        return rotated_vector

    @property
    def cam2lidar_rt(self) -> np.matrix:
        """
        Camera rt matrix relative to radar
        """
        r = Rotation.from_quat(self.heading).as_matrix()
        rt = np.r_[np.c_[r, np.array(self.position)], np.array([[0, 0, 0, 1]])]
        return np.matrix(rt)

    @property
    def intrinsic_matrix(self) -> Optional[np.ndarray]:
        """
        Internal parameter matrix of pinhole and fisheye cameras
        """
        if self.camera_type != CameraType.OmniDirectional.value:
            fx, fy, cx, cy = self.intrinsic
            return np.array([
                [fx, 0, cx],
                [0, fy, cy],
                [0, 0, 1]])
        else:
            return None

    @property
    def _radial(self) -> tuple:
        """
        Transverse distortion
        """
        assert len(self.radial) <= 6, 'The radial distortion of this model only supports k1-k6 at present'
        _ = {f'k{num}': item for num, item in enumerate(self.radial, start=1)}
        k1 = _.get('k1', 0)
        k2 = _.get('k2', 0)
        k3 = _.get('k3', 0)
        k4 = _.get('k4', 0)
        k5 = _.get('k5', 0)
        k6 = _.get('k6', 0)
        return k1, k2, k3, k4, k5, k6

    def _pinhole_3dto2d(self, **kwargs) -> tuple:
        """
        opencv pinhole camera pixel coordinates
        """
        x, y, z = kwargs['point_in_cam']
        fx, fy, cx, cy = self.intrinsic
        k1, k2, k3, k4, k5, k6 = self._radial
        p1, p2 = self.tangential
        u, v = x / z, y / z
        r2 = u ** 2 + v ** 2
        r4 = r2 ** 2
        r6 = r2 * r4
        radial = (1 + k1 * r2 + k2 * r4 + k3 * r6) / (1 + k4 * r2 + k5 * r4 + k6 * r6)
        u = u * radial + 2 * p1 * v * u + p2 * (r2 + 2 * u ** 2)
        v = v * radial + 2 * p2 * v * u + p1 * (r2 + 2 * v ** 2)
        return u * fx + cx, v * fy + cy

    # opencv
    def _fisheye_3dto2d(self, **kwargs) -> tuple:
        """
        opencv fisheye camera pixel coordinates
        """
        x, y, z = kwargs['point_in_cam']
        fx, fy, cx, cy = self.intrinsic
        k1, k2, k3, k4, k5, k6 = self._radial
        alpha = self.skew
        a = x / abs(z)
        b = y / abs(z)
        r2 = a ** 2 + b ** 2
        r = math.sqrt(r2)
        theta = math.atan(r) if z > 0 else math.pi - math.atan(r)

        theta2 = theta ** 2
        theta4 = theta2 ** 2
        theta6 = theta2 * theta4
        theta8 = theta4 ** 2

        theta_d = theta * (1.0 + k1 * theta2 + k2 * theta4 + k3 * theta6 + k4 * theta8)
        scale = 1.0 / r if r > 1e-8 else 1.0

        u, v = theta_d * scale * a, theta_d * scale * b
        return fx * (u + alpha * v) + cx, fy * v + cy

    # kb8
    def _fisheye2_3dto2d(self, **kwargs) -> tuple:
        """
        kb8 fisheye camera pixel coordinates
        """
        x, y, z = kwargs['point_in_cam']
        fx, fy, cx, cy = self.intrinsic
        k1, k2, k3, k4, k5, k6 = self._radial
        alpha = self.skew
        a = x / abs(z)
        b = y / abs(z)
        r2 = a ** 2 + b ** 2
        r = math.sqrt(r2)
        theta = math.atan(r) if z > 0 else math.pi - math.atan(r)

        theta2 = theta ** 2
        theta4 = theta2 ** 2
        theta6 = theta2 * theta4
        theta8 = theta4 ** 2

        theta_d = theta * (1.0 + k1 * theta2 + k2 * theta4 + k3 * theta6 + k4 * theta8)
        scale = 1.0 / r if r > 1e-8 else 1.0

        u, v = theta_d * scale * a, theta_d * scale * b
        return fx * (u + alpha * v) + cx, fy * v + cy

    def _omnidirectional_3dto2d(self, **kwargs) -> tuple:
        """
        opencv omnidirectional camera pixel coordinates
        """
        x, y, z = kwargs['point_in_cam']
        fx, fy, c, d, e, cx, cy = self.intrinsic
        norm = math.sqrt(x * x + y * y)
        theta = math.atan2(-z, norm)
        theta_i, rho = 1, 0
        for pi in self.projection:
            rho += theta_i * pi
            theta_i *= theta
        inv_norm = 1 / norm
        xn1, xn2 = x * inv_norm * rho, y * inv_norm * rho
        return xn1 * c + xn2 * d + cx, xn1 * e + xn2 + cy

    def get_pixel_coordinate(self, point_in_cam: tuple) -> tuple:
        """
        Convert point cloud points to pixel coordinates

        Args:
            point_in_cam: (x, y, z)
                The coordinates of points in the camera coordinate system
        """
        if self.camera_type == CameraType.PinHole.value:
            return self._pinhole_3dto2d(point_in_cam=point_in_cam)
        elif self.camera_type == CameraType.Fisheye.value:
            return self._fisheye_3dto2d(point_in_cam=point_in_cam)
        else:
            return self._omnidirectional_3dto2d(point_in_cam=point_in_cam)

    @staticmethod
    def gen_camera(camera: Dict):
        """
        generate the Camera obj

        Args:
            camera: Dict
                camera data
        Returns:
            obj: Camera
        """

        def _temp_cam(data: Dict):
            if not data:
                return
            if isinstance(data, Dict):
                return tuple(data.values())
            return data

        camera_type = _temp_cam(camera.get("type"))
        assert isinstance(camera_type, str)
        return Camera(camera_type=_temp_cam(camera.get("type")),
                      heading=_temp_cam(camera.get("heading", None)),
                      position=_temp_cam(camera.get("position", None)),
                      intrinsic=_temp_cam(camera.get("intrinsic", None)),
                      radial=_temp_cam(camera.get("radial", None)),
                      tangential=_temp_cam(camera.get("tangential", None)),
                      fov=_temp_cam(camera.get("fov", None)),
                      projection=_temp_cam(camera.get("projection", None))
                      )


if __name__ == '__main__':
    info = {
        'position': {
            'x': 19.617147983589785,
            'y': -0.3571670183003431,
            'z': 0.2083393465328241},
        'scale': {
            'height': 4.793909959547556,
            'width': 2.0032044225562577,
            'depth': 1.5418855556278792},
        'rotation': {
            'x': 0.0,
            'y': 0.0,
            'z': -1.5731558280731583}
    }
    path = '/Users/stardust/Desktop/1681532060081818.pcd'
    points = o3d.io.read_point_cloud(path)
    center = [info['position']['x'], info['position']['y'], info['position']['z']]  # Center coordinates
    extent = [info['scale']['width'], info['scale']['height'], info['scale']['depth']]  # Extent along each axis
    theta = info['rotation']['z']
    rot_mat = np.array([[np.cos(theta), -np.sin(theta), 0],
                        [np.sin(theta), np.cos(theta), 0],
                        [0, 0, 1]])
    bbox = o3d.geometry.OrientedBoundingBox(center=center, R=rot_mat, extent=extent)
    _ = bbox.get_point_indices_within_bounding_box(points.points)
    _len = len(_)
    _ = np.array([np.asarray(points.points)[p] for p in _])
    matrix = np.c_[_, np.array([1] * _len)]
    cam = Camera(
        'PinHole',
        heading=(0.4351436321314605, -0.4677553765588058, 0.5492505771794342, -0.5386824023335351),
        position=(1.5601294, -0.37986112, 1.3333596),
        intrinsic=(1846.008599353862, 1857.930995475229, 1948.09353722228, 1078.768188621484),
        radial=(0, 0, 0, 0),
        tangential=(0, 0)
    )

    import cv2 as cv

    image_path = '/Users/stardust/Desktop/1681532060061245.jpg'
    img = cv.imread(image_path)
    point_size = 1
    point_color = (0, 0, 255)  # BGR
    thickness = 4  # Accepts 0, 4, or 8

    for points in np.dot(matrix, cam.cam2lidar_rt.I.T).tolist():
        point = np.array(cam.get_pixel_coordinate(points[:3]), int)
        cv.circle(img, point, point_size, point_color, thickness)

    cv.namedWindow("image")
    cv.imshow('image', img)
    cv.waitKey(10000)  # Display for 10 seconds
    cv.destroyAllWindows()

    from pathlib import Path

    for file in Path("your_directory").glob("*.json"):
        print(file.name)



import os
from enum import Enum
from retry import retry
from typing import Optional
from dataclasses import dataclass

import requests


class AttachmentType(Enum):
    """
        Media type enumeration
    """
    Text = "TEXT"
    Image = "IMAGE"
    ImageSet = "IMAGE_SET"
    Video = "VIDEO"
    Audio = "AUDIO"
    AudioSet = "AUDIO_SET"
    PointCloud = "POINTCLOUD"
    PointCloudSet = "POINTCLOUD_SET"
    ImageSequence = "IMAGE_SEQUENCE"
    PointCloudSequence = "POINTCLOUD_SEQUENCE"
    PointCloudSetSequence = "POINTCLOUD_SET_SEQUENCE"


@dataclass
class UriInfo:
    uri: str

    @staticmethod
    def _mkdir(path) -> None:
        if not os.path.isfile(path):
            raise ValueError('download_path is error')
        dir_path = os.path.dirname(path)
        os.makedirs(dir_path, exist_ok=True)

    @retry(tries=3)
    def download_url(self, download_path: Optional[str] = None) -> str:
        if download_path:
            save_path = download_path
            self._mkdir(save_path)
        else:
            save_path = f'./{os.path.basename(self.uri)}'
        response = requests.get(self.uri, timeout=3)
        with open(save_path, 'wb') as wp:
            wp.write(response.content)
        return save_path



from typing import List

import numpy as np
import open3d as o3d
from scipy.spatial import ConvexHull
from shapely.geometry import Polygon

from .proj_helper import get_img_points
from stardust.geometry.ops3d.bbox.structure.bbox_np_ops import points_in_rbbox



def obj_occluded_check(points: np.ndarray,
                       img_info: List,
                       save_bboxes: np.ndarray,
                       save_labels: np.ndarray,
                       save_scores: np.ndarray,
                       area_thr=0.7
                       ):
    """
    Check if lidar objects are occluded on img

    Args:
        points:
            point cloud points
        img_info:
            list of dicts contain img and cam info, you can get this through func: ./utils/projection/helper/decode_img_sources
        save_bboxes:
            bounding boxes
        save_labels:
            labels
        save_scores:
            conf scores
        area_thr:
            threshold for occlusion decision

    Return:
        occlusion mask, True means it is occluded
    """
    num_camera = len(img_info)
    box_mask = np.full((save_bboxes.shape[0],), True, dtype=bool)
    # Track distance from origin for each box
    box_distance = np.sqrt(np.sum(save_bboxes[:, :3] ** 2, axis=1))
    save_bboxes_withd = np.hstack((save_bboxes, box_distance.reshape(save_bboxes.shape[0], 1)))  # N * 4
    # Sort boxes by their distance to the origin
    sorted_indices = np.argsort(save_bboxes_withd[:, -1])
    save_bboxes_withd = save_bboxes_withd[sorted_indices]
    save_bboxes = save_bboxes_withd[:, :-1]
    save_labels = save_labels[sorted_indices]
    save_scores = save_scores[sorted_indices]
    points_mask = points_in_rbbox(points, save_bboxes)
    # Maintain polygons already projected per image
    img_bbox = {i: [] for i in range(num_camera)}
    model_car = o3d.io.read_point_cloud('./car_model_modified.pcd')
    model_car = np.asarray(model_car.points)
    l_model_car = np.max(model_car[:, 0]) - np.min(model_car[:, 0])
    w_model_car = np.max(model_car[:, 1]) - np.min(model_car[:, 1])
    h_model_car = np.max(model_car[:, 2]) - np.min(model_car[:, 2])
    for m in range(points_mask.shape[1]):
        obj_pts = points[points_mask[:, m]]
        if len(obj_pts) <= 200:
            box_mask[m] = False
            continue
        obj_box = save_bboxes[m].tolist()
        x, y, z, width, length, height, ry = obj_box
        scale = np.array([length / l_model_car, width / w_model_car, height / h_model_car])
        model_car_scaled = model_car * scale
        gamma = - np.pi / 2 - ry
        rm = np.array([[np.cos(gamma), -np.sin(gamma), 0],
                       [np.sin(gamma), np.cos(gamma), 0],
                       [0, 0, 1]])
        model_car_scaled = np.matmul(model_car_scaled, rm.T)
        model_car_scaled[:, [0, 1, 2]] = model_car_scaled[:, [0, 1, 2]] + np.array([x, y, z])
        obj_pts_backup, obj_pts = obj_pts, model_car_scaled
        points_proj_res = get_img_points(obj_pts, num_camera, img_info)
        points_proj_res_backup = get_img_points(obj_pts_backup, num_camera, img_info)
        camera_mask = []  # Track visibility per camera
        for i, (x, y, h, w, occluded_pts) in points_proj_res.items():
            # Iterate cameras and check if the projection stays in frame
            proj_pixels = []
            for x_, y_ in zip(x, y):
                if 0 <= x_ < w and 0 <= y_ < h:
                    proj_pixels.append([x_, y_])
            proj_pixels = np.array(proj_pixels)  # object pixels on image
            if i in points_proj_res_backup:
                proj_pixels_backup = []
                val = points_proj_res_backup[i]
                x, y, h, w = val[0], val[1], val[2], val[3]
                for x_, y_ in zip(x, y):
                    if 0 <= x_ < w and 0 <= y_ < h:
                        proj_pixels_backup.append([x_, y_])
                proj_pixels_backup = np.array(proj_pixels_backup)  # object pixels on image
            if occluded_pts == 0:
                # Projection fully visible with no occlusion samples
                hull = ConvexHull(proj_pixels)
                hull1 = hull.vertices.tolist()  # Close polygon by returning to the first vertex
                hull1.append(hull1[0])
                poly_vertices = proj_pixels[hull1, :]
                obj_2d_cur = Polygon(poly_vertices)
                hull_backup = ConvexHull(proj_pixels_backup)
                hull1 = hull_backup.vertices.tolist()  # Close polygon by returning to the first vertex
                hull1.append(hull1[0])
                poly_vertices = proj_pixels_backup[hull1, :]
                obj_2d_cur_backup = Polygon(poly_vertices)
                # Compare observed area against the scaled model area first
                if obj_2d_cur_backup.area / obj_2d_cur.area <= (1 - area_thr):
                    camera_mask.append(False)
                masked = True  # Track whether the box is occluded
                if img_bbox[i]:
                    # Another closer box is already tracked; compute IoU
                    whole_union = None
                    for obj_2d_pre in img_bbox[i]:
                        if (obj_2d_pre).intersection(obj_2d_cur).area > 0:
                            if whole_union is None:
                                whole_union = (obj_2d_pre).intersection(obj_2d_cur)
                            else:
                                whole_union = whole_union.union((obj_2d_pre).intersection(obj_2d_cur))
                    if whole_union is not None:
                        masked_area = whole_union.area / (obj_2d_cur).area
                        if masked_area >= area_thr:
                            masked = False
                    camera_mask.append(masked)
                    img_bbox[i].append(obj_2d_cur)
                else:
                    img_bbox[i].append(obj_2d_cur)
                    camera_mask.append(masked)
                # Truncated projection path
            else:
                if len(proj_pixels) >= 3:
                    hull = ConvexHull(proj_pixels)
                    hull1 = hull.vertices.tolist()  # Close polygon by returning to the first vertex
                    hull1.append(hull1[0])
                    poly_vertices = proj_pixels[hull1, :]
                    obj_2d_cur = Polygon(poly_vertices)
                    iou_flag = True
                    masked = True  # Track whether the box is occluded
                    if img_bbox[i]:
                        # Another closer box is already tracked; compute IoU
                        for obj_2d_pre in img_bbox[i]:
                            intersect = (obj_2d_pre).intersection(obj_2d_cur).area
                            union = (obj_2d_pre).union(obj_2d_cur).area
                            masked_area = intersect / (obj_2d_cur).area
                            if masked_area >= area_thr:
                                masked = False
                                break
                        camera_mask.append(masked)
                        img_bbox[i].append(obj_2d_cur)
                        iou_flag = False
                    else:
                        img_bbox[i].append(obj_2d_cur)
                    if iou_flag:
                        if (1 - area_thr) * obj_pts.shape[0] > occluded_pts:
                            camera_mask.append(True)
                        else:
                            camera_mask.append(False)
                else:
                    camera_mask.append(False)
        camera_mask = np.any(camera_mask)
        if not camera_mask:
            box_mask[m] = False
    return box_mask



import os
import re
import json
import asyncio
from pathlib import Path
from loguru import logger
from urllib.parse import quote, unquote
from typing import List, Sequence, Dict, Tuple

import aiohttp
import uvloop
from tqdm import tqdm
from stardust.components.attachment.image import ALL_IMAGE
from stardust.components.attachment.pointcloud import ALL_POINTCLOUD
from stardust.rosetta.rosetta_data import RosettaData
from stardust.components.frame import Frame

asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
logger.add("app.log", rotation="500 MB")


class Downloader:
    def __init__(self, urls: List[Tuple[str]], save_path: str = None):
        """
        downloader

        Args:
            urls:
                Required download link
            save_path:
                Save directory
        """
        self.urls = urls
        self.save_path = save_path
        self.slice_num = 1000
        self.attempts = 3
        self.bar = tqdm(total=len(self.urls))
        self.err_list = list()

    @staticmethod
    def is_empty(filename):
        return os.path.getsize(filename) != 0

    @staticmethod
    async def write_local(result, local_path):
        os.makedirs(os.path.dirname(local_path), exist_ok=True)
        with open(local_path, 'wb') as wp:
            wp.write(result)
            return True

    async def fetch(self, session, url, local_path):
        for n in range(self.attempts):
            try:
                async with session.get(url) as response:
                    if response.status != 200:
                        print(f"Failed to download {url}  status_code:{response.status}")
                        continue
                    content = await response.content.read()
                    result = await self.write_local(content, local_path)
                    if result:
                        self.bar.update(1)
                        if n != 0:
                            print(f'{n} Retry successfully {url}')
                        return True
            except Exception as e:
                print(f'Failed to download {url}: {e}')
            await asyncio.sleep(0.3)
        self.err_list.append(f"Failed to download {url}")
        return False

    async def async_download(self):
        async with aiohttp.ClientSession() as session:
            for i in range(0, len(self.urls), self.slice_num):
                tasks = []
                for url, file_save_path in self.urls[i: i + self.slice_num]:
                    url = quote(url, safe=':/')
                    if os.path.exists(file_save_path) and self.is_empty(file_save_path):
                        self.bar.update(1)
                        continue
                    tasks.append(asyncio.create_task(self.fetch(session, url, file_save_path)))
                if tasks:
                    await asyncio.wait(tasks)
            if self.err_list:
                error_txt = os.path.join(os.getcwd(), 'download_error.txt')
                with open(error_txt, 'w') as wp:
                    for _ in self.err_list:
                        wp.write(_)
                        wp.write("\n")
                    print(f'{len(self.err_list)} urls failed to download')
            self.bar.close()


def attachment_name_factory(att, save_path, project_id, task_id, frame_num, image_url_lst, pointcloud_url_lst,
                            name=None):
    if att.startswith("oss://stardust-data/"):
        att = att.replace("oss://stardust-data/", "https://stardust-data.oss-cn-hangzhou.aliyuncs.com/")
    if (suffix := Path(att).suffix) in ALL_IMAGE:
        if name:
            image_url_lst.append(
                (att, os.path.join(save_path, str(project_id), "images", name, f"{task_id}_{frame_num}{suffix}")))
        else:
            image_url_lst.append(
                (att, os.path.join(save_path, str(project_id), "images", f"{task_id}_{frame_num}{suffix}")))
    elif suffix in ALL_POINTCLOUD:
        pointcloud_url_lst.append(
            (att, os.path.join(save_path, str(project_id), "pcds", f"{task_id}_{frame_num}{suffix}")))
    else:
        raise TypeError("Dataset type not supported yet")


def load_dataset(frame: Frame = None,
                 project_id: int = None,
                 save_path=None,
                 **kwargs) -> None:
    """
    The function of loading data sets, you can enter the Frame object, you can also enter the project pool, if there is already a local rosetta_json will not be downloaded again
    save_path is used as the path to save annotation files. If the path has a frame, the data in the frame is obtained. Otherwise, the project_id,, and other values are required.

    Args:
        frame:
            Frame object
        save_path:
            Save directory
        peoject_id:
            rosetta project ID
        pool_lst:
            pool list

    Returns:

    Outputs:
        path
            -{project_id}
                --jsons
                    {task_id}_{frame_id}.json
                --images
                    --{camera_id}
                        {task_id}_{frame_id).jpg
                --pcds
                    {task_id}_{frame_id}.pcd
                --poses
                    {task_id}_{frame_id).txt
                --calibrations
                    {camera_id}.json

    """
    assert save_path
    if save_path and not os.path.exists(save_path):
        os.makedirs(save_path)

    # Links that need to be downloaded
    image_url_lst = list()
    pointcloud_url_lst = list()
    # Handle in-memory Frame objects
    if frame:
        project_id = frame.task_info.project_id  # Rosetta project ID
        task_id = frame.task_info.task_id  # Task ID
        frame_num = frame.task_info.frame_num  # Frame index
        media = frame.media  # Annotation files
        # Check for point clouds first
        if hasattr(media, "point_cloud") and media.point_cloud and media.point_cloud.uri:
            attachment_name_factory(media.point_cloud.uri, save_path, project_id, task_id, frame_num, image_url_lst,
                                    pointcloud_url_lst)
        # Download related images if present
        if hasattr(media, "image") and media.image:
            if isinstance(media.image, Sequence):
                for img in media.image:
                    attachment_name_factory(img.uri, save_path, project_id, task_id, frame_num, image_url_lst,
                                            pointcloud_url_lst, name=img.camera_param.name)
            else:
                attachment_name_factory(media.image.uri, save_path, project_id, task_id, frame_num, image_url_lst,
                                        pointcloud_url_lst)
    # Or fetch directly from a project ID
    elif project_id is not None:
        assert isinstance(project_id, int)
        pool_lst = kwargs.get("pool_lst", [])  # Pool IDs for the Rosetta project
        RosettaData(project_id, save_path).export(pool_lst)

        # Prepare directory for processed JSON files
        jsons_save_path = os.path.join(save_path, str(project_id), "jsons")
        os.makedirs(jsons_save_path, exist_ok=True)
        # Iterate Rosetta JSON to extract download links
        for file in Path(save_path).joinpath(str(project_id), "json").glob(f"*.json"):
            with open(file, 'r', encoding='utf-8') as f:
                attachment = (data := json.load(f))['taskParams']['record']['attachment']
                task_id = data['taskId']
                # Frame number
                frame_num = str(file).rsplit(".", 1)[0][-4:]
                if not frame_num.isdigit():
                    frame_num = '0001'
                # Extract URLs and destinations
                if isinstance(attachment, Dict):
                    # re_find_lst = re.findall("'url': '(?P<urls>.*?)',", str(attachment))
                    main_url = attachment['url']
                    attachment_name_factory(main_url, save_path, project_id, task_id, frame_num, image_url_lst,
                                            pointcloud_url_lst)
                    for att in attachment.get("imageSources", []):
                        img_url = att['url']
                        name = att['name']
                        attachment_name_factory(img_url,
                                                save_path,
                                                project_id,
                                                task_id,
                                                frame_num,
                                                image_url_lst,
                                                pointcloud_url_lst, name)
                elif isinstance(attachment, str):
                    attachment_name_factory(attachment,
                                            save_path,
                                            project_id,
                                            task_id,
                                            frame_num,
                                            image_url_lst,
                                            pointcloud_url_lst)
            with open(f'{jsons_save_path}/{task_id}_{frame_num}.json', 'w', encoding='utf-8') as new_json:
                json.dump(data, new_json, ensure_ascii=False)
    else:
        raise TypeError("Missing required parameters")

    asyncio.run(Downloader(pointcloud_url_lst).async_download())
    asyncio.run(Downloader(image_url_lst).async_download())
    return save_path


if __name__ == '__main__':
    load_dataset(save_path="/Users/mac/Documents/pyproject/sd_sdk/stardust_sdk/Stardust_SDK/data/",
                 project_id=1529, pool_lst=[])



import os, shutil
from setuptools import setup, find_packages
from distutils.cmd import Command

# Remove the build directory produced during packaging
# CUR_PATH = os.path.dirname(os.path.abspath(__file__))
# path = os.path.join(CUR_PATH, 'build')
# if os.path.isdir(path):
#     print('INFO del dir ', path)
#     shutil.rmtree(path)

import configparser


def generate_config_file():
    # Create a ConfigParser instance
    config = configparser.ConfigParser()

    # Populate configuration sections
    config['OssUser'] = {
        'key': 'LTAI4GHWdWdjGDE4Gx1Efbyt',
        'secret': 'SffMd625UN8ceXLj0RlRG8oyF8BkjU',
        'bucket': 'stardust-data',
        'endpoint': 'https://oss-cn-hangzhou.aliyuncs.com',
        'prefix_url': 'https://stardust-data.oss-cn-hangzhou.aliyuncs.com'
    }

    config['RosUser'] = {
        'username': '13693170335',
        'password': 'Start123!'
    }

    config['RosConfig'] = {
        'env_prod': 'https://server.rosettalab.top',
        'env_dev': 'https://dev-server.rosettalab.top'
    }

    # Write configuration to disk
    with open('config222.ini', 'w') as configfile:
        config.write(configfile)


class CustomCommand(Command):
    description = 'Custom command description'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        print("Executing custom command...")  # Custom command implementation
        # Create a ConfigParser instance
        config = configparser.ConfigParser()

        # Populate configuration sections
        config['OssUser'] = {
            'key': 'LTAI4GHWdWdjGDE4Gx1Efbyt',
            'secret': 'SffMd625UN8ceXLj0RlRG8oyF8BkjU',
            'bucket': 'stardust-data',
            'endpoint': 'https://oss-cn-hangzhou.aliyuncs.com',
            'prefix_url': 'https://stardust-data.oss-cn-hangzhou.aliyuncs.com'
        }

        config['RosUser'] = {
            'username': '13693170335',
            'password': 'Start123!'
        }

        config['RosConfig'] = {
            'env_prod': 'https://server.rosettalab.top',
            'env_dev': 'https://dev-server.rosettalab.top'
        }

        # Write configuration to disk
        with open('config222.ini', 'w') as configfile:
            config.write(configfile)


# Call helper to generate the config file
# generate_config_file()


setup(
    name='stardust-sdk',  # Package name - changed to avoid conflict with existing 'stardust' package
    author='stardust.ai',
    author_email='support@stardust.ai',
    version='0.1.1',  # Package version - using semantic versioning
    description='Stardust SDK for AI/ML data processing and annotation workflows',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/stardustai/Stardust-SDK',
    project_urls={
        'Bug Reports': 'https://github.com/stardustai/Stardust-SDK/issues',
        'Source': 'https://github.com/stardustai/Stardust-SDK',
        'Documentation': 'https://sdk-docs.stardust.ai/',
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    keywords='ai ml computer-vision annotation data-processing stardust',
    packages=find_packages(),
    include_package_data=True,
    package_data={
        'stardust': ['config.ini', "visualization/base/*.json"],
    },
    entry_points={
        'console_scripts': [
            'stardust = stardust.command.export:main',
        ],
    },
    install_requires=[
        'numpy>=1.24.0',
        'opencv-python>=4.8.0',
        'open3d>=0.17.0',
        'requests>=2.31.0',
        'matplotlib>=3.7.0',
        'pillow>=10.0.0',
        'aiohttp>=3.9.0',
        'httpx>=0.27.0',
        'jsonlines>=4.0.0',
        'numba>=0.58.0',
        'pandas>=2.1.0',
        'retry>=0.9.0',
        'scipy>=1.11.0',
        'shapely>=2.0.0',
        'tqdm>=4.66.0',
        'uvloop>=0.19.0',
    ],
    python_requires=">=3.9",
    extras_require={
        'dev': [
            'pytest>=7.0',
            'pytest-cov>=4.0',
            'black>=23.0',
            'flake8>=6.0',
        ],
    },
)



import numpy as np


class Segmentation:
    def __repr__(self) -> str:
        description = 'Calculate segmentation metric'
        return description

    def __init__(self, gts: np.array, preds: np.array, num_classes: int) -> None:
        self.num_classes = num_classes
        assert np.max(gts) < self.num_classes
        assert np.max(preds) < self.num_classes
        assert len(preds.size()) == 2 and len(gts.size()) == 2
        self.gts = gts
        self.preds = preds
        self.h, self.w = self.gts.shape
        self.confuse_matrix = self.confusion_matrix()

    def confusion_matrix(self) -> np.array:
        """
        compute confusion matrix

        Returns:
            confusion_matrix(ndarray):
        """
        results = np.zeros((self.num_classes, self.num_classes), dtype=np.uint8)
        for i, j in zip(self.gts.flatten(), self.preds.flatten()):
            results[i, j] += 1
        return results

    def global_pixel_accuracy(self):
        """
        compute global pixel accuracy

        Returns:
            global pixel accuracy(float):
        """
        return len(np.where(self.gts == self.preds)[0]) / (self.h * self.w)

    def class_accuracy(self, class_id: int):
        """
        compute accuracy of the ith class

        Args:
            class_id: int

        Returns:
            class accuracy: float
        """
        assert class_id >= self.num_classes
        return self.confuse_matrix[class_id, class_id] / np.sum(self.confuse_matrix[class_id, :])

    def mean_pixel_accuracy(self):
        """
        compute accuracy of all classes

        Returns:
            mean pixel accuracy: float
        """
        result = 0
        for class_id in range(self.num_classes):
            result += (self.confuse_matrix[class_id, class_id] / np.sum(self.confuse_matrix[class_id, :]))
        return result / self.num_classes

    def IOU(self, class_id: int):
        """
        compute IOU of the ith class

        Args:
            class_id: int

        Returns:
            IOU: float
        """
        assert class_id >= self.num_classes
        return self.confuse_matrix[class_id, class_id] / (np.sum(self.confuse_matrix[class_id, :]) + np.sum(self.confuse_matrix[:, class_id]) - self.confuse_matrix[class_id, class_id])

    def mean_IOU(self):
        """
        compute mean IOU of all classes

        Returns:
            mean IOU: float
        """
        result = 0
        for class_id in range(self.num_classes):
            result += self.confuse_matrix[class_id, class_id] / (np.sum(self.confuse_matrix[class_id, :]) + np.sum(self.confuse_matrix[:, class_id]) - self.confuse_matrix[class_id, class_id])
        return result / self.num_classes



import os
import json
from pathlib import Path

from stardust.convertion.to_coco.coco_demo import COCO_ANNS
from stardust.utils.convert import read_rosetta

_all_input = set()
_input_parent_map = dict()
_input_value_id_map = dict()
_g_input_id = 1
_g_box_id = 1


# COCO demo scaffold
def _coco_demo():
    return COCO_ANNS


def func_category(input: "sdk input"):
    if input.input_type == "select":
        return input.value
    elif input.input_type == "nested":
        return [input.value[0], input.value[1]]


# Convert Stardust data into COCO structure
def sd_to_coco(demo: "coco format",
               data: "sdk format",
               frame_num,
               ensure_match=False):
    global _g_input_id
    global _g_box_id

    demo["images"].append({
        "license": 1,
        "file_name": Path((image := data.media.image).uri).name,
        "coco_url": image.uri.replace("oss://stardust-data/Clients/", "https://stardust-data.oss-cn-hangzhou.aliyuncs.com/Clients/"),
        "height": 2160,
        "width": 3840,
        "date_captured": None,
        "flickr_url": None,
        "id": frame_num + 1
    })

    for index, (_id, ann) in enumerate(data.annotation.input_lst.items()):
        # Skip fields without a value
        if not ann.value:
            continue

        # Keep only category-style inputs
        if ann.name not in ("category", "Type", "type", "portrait_id"):
            continue

        # Map box2d entries back to their input field (1:1 mapping)
        category_id = int(f"{frame_num}{index}")
        if ann.parent:
            _input_parent_map.setdefault(ann.parent, (_id, category_id))

        # Category value
        assert ann
        category = func_category(ann)

        # Only register each value once
        if category in _all_input:
            continue
        _all_input.add(category)
        _input_value_id_map.setdefault(category, _g_input_id)

        # Append category definition
        demo["categories"].append({
            "supercategory": category,
            "id": _g_input_id,
            "name": category,
            # "origin_id": ann.id
        })

        _g_input_id += 1

    for index, (_id, ann) in enumerate(data.annotation.box2d_lst.items()):
        if ensure_match and not _input_parent_map.get(_id):
            continue

        # Retrieve the input annotation object
        _input_obj = data.annotation.input_lst.get(_input_parent_map[_id][0])
        if not _input_obj:
            continue
        category = func_category(_input_obj)

        _children = ann.children
        segmentation = list()
        area = ann.area
        if polygon_lst := data.annotation.polygon_lst:
            for _child in _children:
                if polygon_lst.get(_child):
                    _segem = list()
                    points = polygon_lst[_child].points
                    segmentation = list()
                    for i in points:
                        _segem.extend([round(i[0], 2), round(i[1], 2)])
                    area = polygon_lst[_child].area
                    segmentation.append(_segem)
        demo["annotations"].append({
            "segmentation": segmentation,
            "area": area,
            "iscrowd": 0,
            "image_id": frame_num + 1,
            "bbox": [round(ann.x1, 2), round(ann.y1, 2), ann.width, ann.height],
            "category_id": _input_value_id_map[category] if _input_parent_map.get(_id) else None,
            # "id": int(f"{frame_num}{index}")
            "id": _g_box_id,
            # "origin_id": ann.id
        })
        _g_box_id += 1


# Stardust exports yield a generator
def coco_export(sd_generator: "sdk generator",
                save_path: "coco output path"):
    # COCO demo payload
    demo = _coco_demo()

    # Convert Stardust frames to COCO annotations
    for index, json_data in enumerate(sd_generator):
        sd_to_coco(demo, json_data, index, ensure_match=True)

    # Persist COCO annotations
    if os.path.exists(save_path):
        os.remove(save_path)
    with open(save_path, 'w', encoding='utf-8') as sf:
        json.dump(demo, sf, ensure_ascii=False)

    print(f"Exported COCO file saved to: {save_path}")


if __name__ == '__main__':
    # Example usage:

    # Data directory
    base_dir = "/Users/mac/Documents/pyproject/sd_sdk/stardust_sdk/Stardust_SDK/data"
    # Rosetta project ID
    project_id = 2066

    # Prepare directories
    if not os.path.exists(base_dir):
        os.makedirs(base_dir)
    input_path = f"{base_dir}/{project_id}"

    # Convert from Rosetta to Stardust
    sd_generator = read_rosetta(project_id=project_id,
                                input_path=input_path,
                                )

    # COCO output path
    os.makedirs(f"{base_dir}/{project_id}/output/", exist_ok=True)
    res_save_path = f"{base_dir}/{project_id}/output/{project_id}_coco.json"

    # Run conversion
    coco_export(sd_generator, res_save_path)



from .audio import Audio
from .point import Point
from .point3d import Point3D
from .line import Line
from .polygon import Polygon
from .box import Box2D
from .box3d import Box3D
from .text import Text
from .video import Video
from .input import Input



from typing import List, Union
from numpy import array
from .point3d import Point3D
from .base_annotation import AnnotationObject


class Segementation(AnnotationObject):
    """
    3d semantic segmentation
    """

    def __init__(self, points: Union[List[Point3D], array], **kwargs):
        self.points = points
        super().__init__(**kwargs)



from .base_annotation import AnnotationType, AnnotationObject


class Video(AnnotationObject):
    def __init__(self,
                 start,
                 end,
                 fps=30,
                 **kwargs):
        """
        This class defines the structure of video label task

        Args:
           start:
               The start time of annotation for this text
           end:
               The end time of annotation for this text

        Returns:
           Video class

        Examples:
            .. code-block:: python

                video = Video(start=0.57, end=12.13)
        """
        assert isinstance(start, float)
        assert isinstance(end, float)
        assert isinstance(fps, int)
        self.start = start
        self.end = end
        self.fps = fps
        AnnotationObject.__init__(self, **kwargs)
